# pylint: disable=E1101
"""
@file
@brief Gradient with :epkg:`onnxruntime-training` forward backward.
"""
import os
import logging
import warnings
from io import BytesIO
import onnx
from onnx.numpy_helper import to_array
from onnxruntime import InferenceSession, RunOptions
from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611
    SessionIOBinding, OrtValue as C_OrtValue)
from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611
    TrainingAgent, OrtValueCache, OrtModuleGraphBuilder,
    OrtModuleGraphBuilderConfiguration, OrtDevice,
    TrainingGraphTransformerConfiguration, OrtValueVector,
    PartialGraphExecutionState)
from ..utils.orttraining_helper import get_train_initializer


class OrtGradientForwardBackward:
    """
    Implements forward backward mechanism assuming the function
    to train is defined by an ONNX graph.

    :param onnx_model: onnx model
    :param weights_to_train: names of the weights to train,
        if None, all initializer of floats type are included in the list
    :param input_names: input names or None for all
    :param output_names: output names or None for all
    :param class_name: name to give the class dynamically created
    :param sess_options: see :epkg:`SessionOptions`
    :param providers: see :epkg:`InferenceSession`
    :param provider_options: see :epkg:`InferenceSession`
    :param run_options: see :epkg:`RunOptions`
    :param graph_builder_config:
        see :epkg:`OrtModuleGraphBuilderConfiguration`
    :param device_index: used for cuda (0 for `cuda:0`,
        `cuda:1`, ...), 0 by default
    :param enable_logging: enables logging while setting up the class
    :param debug: to run extra verification while training

    .. note::
        The current implementation of :epkg:`onnxruntime` forces
        the weights to train to appear in the alphabetical order.
        The constructor checks that condition is verified.

    .. warning::
        This class does not consider subgraphs.
    """

    def __init__(self, onnx_model, weights_to_train=None,
                 input_names=None, output_names=None, class_name=None,
                 sess_options=None, providers=None,
                 provider_options=None, run_options=None,
                 graph_builder_config=None,
                 device_index=0, enable_logging=False, debug=False):

        if weights_to_train is None:
            weights_to_train = (
                OrtGradientForwardBackward._select_initializer_names(
                    onnx_model))
            if len(weights_to_train) == 0:
                raise RuntimeError(  # pragma: no cover
                    "Unable to guess the weights to train from initializers: "
                    "%r." % [i.name for i in onnx_model.graph.initializer])

        self.onnx_model = onnx_model
        self.input_names = input_names
        self.output_names = output_names
        self.weights_to_train = weights_to_train
        self.device_index = device_index
        self.enable_logging = enable_logging
        self.class_name = (class_name if class_name is not None else
                           "OrtGradientForwardBackwardFunction_%d" % id(self))

        self.provider_options = provider_options
        self.sess_options = sess_options
        self.providers = providers
        self.run_options = run_options
        self.graph_builder_config = graph_builder_config
        self.debug = debug

        # default
        if self.weights_to_train is None:
            raise ValueError(  # pragma: no cover
                "weights_to_train must be specified.")
        if self.input_names is None:
            self.input_names = [obj.name
                                for obj in self.onnx_model.graph.input]
        if self.output_names is None:
            self.output_names = [obj.name
                                 for obj in self.onnx_model.graph.output]
        if self.class_name is None:
            self.class_name = "TorchOrtFunction_%r" % id(
                self)  # pragma: no cover
        if hasattr(self.providers, 'type'):
            if self.providers.type != 'cpu':
                self.device_index = self.providers.index
            self.providers = self.providers.type
        if self.providers in (None, 'cpu'):
            self.providers = ["CPUExecutionProvider" for i in self.input_names]
            if self.provider_options is None:
                self.provider_options = [{} for i in self.input_names]
        elif self.providers in ('cuda', 'cuda:0', 'gpu'):
            self.providers = [
                "CUDAExecutionProvider" for i in self.input_names]
            if self.provider_options is None:
                self.provider_options = [{} for i in self.input_names]
        if len(self.input_names) != len(self.providers):
            raise ValueError(  # pragma: no cover
                "input_names and providers must have the same length.")
        if self.provider_options is None:
            self.provider_options = [{} for i in self.input_names]
        if len(self.input_names) != len(self.provider_options):
            raise ValueError(  # pragma: no cover
                "input_names and provider_options must have the same length.")

        if list(sorted(self.weights_to_train)) != self.weights_to_train:
            raise ValueError(  # pragma: no cover
                "List of weights to train must be sorted but %r is not. "
                "You shoud use function onnx_rename_weights to do that "
                "before calling this class." % self.weights_to_train)
        set_weights = set(self.weights_to_train)
        if len(set_weights) != len(self.weights_to_train):
            raise ValueError(  # pragma: no cover
                "One weight is not unique in %r." % self.weights_to_train)
        found = []
        for i in self.onnx_model.graph.initializer:
            if i.name not in set_weights:
                continue
            found.append(i.name)
        if len(found) != len(self.weights_to_train):
            raise ValueError(
                "One weight name in self.weights_to_train was not found in "
                "the initializers %r found=%r init names=%r." % (
                    self.weights_to_train, found,
                    [i.name for i in self.onnx_model.graph.initializer]))
        if found != self.weights_to_train:
            raise ValueError(
                "List of weights to train must be sorted and follow the "
                "as the initializers in the graph. %r != %r."
                "You shoud use function onnx_rename_weights to do that "
                "before calling this class." % (
                    self.weights_to_train, found))

        if any(map(lambda v: v not in ['CPUExecutionProvider',
                                       'CUDAExecutionProvider'],
                   self.providers)):
            raise ValueError(
                "Unexpected providers %r (providers=%r)." % (
                    self.providers, providers))

        # complete initialisation
        self._init_next()

    @staticmethod
    def _select_initializer_names(onnx_model):
        """
        Selects all initializers with float type.

        :param onnx_model: ONNX graph
        """
        inits = get_train_initializer(onnx_model)
        return list(inits)

    def _init_next(self):
        if self.enable_logging:
            self._logger = logging.getLogger("onnxcustom")
        else:
            self._logger = None  # pragma: no cover
        if self.run_options is None:
            self.run_options = RunOptions()
            self.run_options.training_mode = True

        if self.graph_builder_config is None:
            initializer_names = [
                i.name for i in self.onnx_model.graph.initializer]
            input_names = [i.name for i in self.onnx_model.graph.input]

            config = OrtModuleGraphBuilderConfiguration()
            config.initializer_names = [init for init in initializer_names
                                        if init in self.weights_to_train]
            config.initializer_names_to_train = self.weights_to_train
            config.input_names_require_grad = input_names
            config.build_gradient_graph = True

            if (len(config.initializer_names) !=  # noqa
                    len(config.initializer_names_to_train)):
                raise RuntimeError(  # pragma: no cover
                    "Unable to automatically fill "
                    "OrtModuleGraphBuilderConfiguration, mismatch between "
                    "%r and %r (initializer_names=%r)." % (
                        config.initializer_names,
                        config.initializer_names_to_train,
                        initializer_names))

            p = TrainingGraphTransformerConfiguration()
            config.graph_transformer_config = p

            # config.enable_caching = True
            # config.loglevel =
            # config.use_memory_efficient_gradient = True
            self.graph_builder_config = config

        attributes = self._create_onnx_graphs()
        attributes['__doc__'] = (
            "Inherits from @see cl OrtGradientForwardBackwardFunction.")
        attributes['__module__'] = (
            OrtGradientForwardBackwardFunction.__module__)
        self.cls_type_ = type(
            self.class_name, (OrtGradientForwardBackwardFunction,),
            attributes)

    def new_instance(self):
        """
        Creates an instance of class `self.cls_type_`.
        It implements methods *forward* and *backward*.
        """
        return self.cls_type_()

    def __getstate__(self):
        "Removes any non pickable attribute."
        atts = [k for k in self.__dict__ if not k.endswith('_')
                if k not in {'_logger', 'graph_builder_config',
                             'run_options'}]
        state = {att: getattr(self, att) for att in atts}
        state['run_options'] = None
        state['graph_builder_config'] = None
        return state

    def __setstate__(self, state):
        "Restores any non pickable attribute."
        for att, v in state.items():
            setattr(self, att, v)
        self._init_next()
        return self

    def __repr__(self):
        "usual"
        return "%s(...)" % self.__class__.__name__

    @staticmethod
    def _repr_helper_(obj, indent=0):
        "used to improve logging messages"
        if obj is None:
            return 'None'
        rows = []
        for c in sorted(dir(obj)):
            if c[0] == '_':
                continue
            try:
                value = getattr(obj, c)
            except AttributeError:  # pragma: no cover
                continue
            rows.append("%s=%r" % (c, value))

        if indent == 0:
            return "%s(%s)" % (obj.__class__.__name__, ", ".join(rows))
        return "%s(\n    %s)" % (
            obj.__class__.__name__,
            "\n    ".join(rows))

    @staticmethod
    def _provider_name_to_device_type(provider_name):
        if provider_name == 'CPUExecutionProvider':
            return OrtDevice.cpu()
        if provider_name == 'CUDAExecutionProvider':  # pragma: no cover
            return OrtDevice.cuda()
        raise ValueError(  # pragma: no cover
            'Unexpected provider name %r.' % provider_name)

    def get_initializer(self, name, exc=True):
        """
        Returns an initializer as numpy arrays.

        :param name: initializer name
        :param exc: raises an exception if not found or return None
        :return: the initializer as a :epkg:`C_OrtValue`
        """
        for init in self.onnx_model.graph.initializer:
            if name == init.name:
                return to_array(init)
        if exc:
            raise RuntimeError(  # pragma: no cover
                "Unable to find name %r in %r." % (
                    name,
                    list(i.name for i in self.onnx_model.graph.initializer)))
        return None

    def _create_onnx_graphs(self):
        """
        Creates forward and backward ONNX graph.
        The new class has the following attributes:

        * `__doc__`: doc string
        * `__module__`: module name (this file)
        * `_run_options`: see :epkg:`RunOptions`
        * `_sess`: :epkg:`InferenceSession` with the original graph
        * `_sess_eval`: :epkg:`InferenceSession` on the graph
            with weights as inputs
        * `_training_agent`: :epkg:`TrainingAgent`
        * `_cache`: :epkg:`OrtValueCache`
        * `_logger`: logger
        * `_input_names`: input names
        * `_debug`: use debug mode
        * `_grad_input_names`: gradient input names
        * `_output_names`: output names
        * `_weights_to_train`: names of the weights to train

        Training attributes

        * `_bw_fetches_names`: bw_fetches_names,
        * `_fw_outputs_device_info`: fw_outputs_device_info,
        * `_bw_outputs_device_info`: bw_outputs_device_info,
        * `_fw_no_grad_output_device_info`: fw_no_grad_output_device_info,
        * `_graph_info`: graph_info}

        Additional attributes added if *keep_model* is True:

        * `_trained_onnx`: ONNX graph for the gradient
        * `_optimized_pre_grad_model`: evaluation ONNX graph taking
            weights as inputs
        * `_graph_builder`: :epkg:`OrtModuleGraphBuilder`
        """
        logger = self._logger
        if logger is not None:
            logger.info("[OrtGradientForwardBackward] create training onnx")
            logger.info("[OrtGradientForwardBackward] input_names=%r",
                        self.input_names)
            logger.info("[OrtGradientForwardBackward] output_names=%r",
                        self.output_names)
            logger.info("[OrtGradientForwardBackward] weights_to_train=%r",
                        self.weights_to_train)

        builder = OrtModuleGraphBuilder()

        if logger is not None:
            cf = self.graph_builder_config.graph_transformer_config
            cfp = cf.propagate_cast_ops_config
            logger.info(
                "[OrtGradientForwardBackward] "
                "OrtModuleGraphBuilder.initialize")
            logger.info(
                "[OrtGradientForwardBackward] graph_builder_config=%s",
                OrtGradientForwardBackward._repr_helper_(
                    self.graph_builder_config, indent=4))
            logger.info(
                "[OrtGradientForwardBackward] graph_builder_config."
                "graph_transformer_config=%s",
                OrtGradientForwardBackward._repr_helper_(cf, indent=4))
            logger.info(
                "[OrtGradientForwardBackward] graph_builder_config."
                "graph_transformer_config.propagate_cast_ops_config=%s",
                OrtGradientForwardBackward._repr_helper_(cfp, indent=4))

        builder.initialize(
            self.onnx_model.SerializeToString(),
            self.graph_builder_config)

        if logger is not None:
            logger.info(
                "[OrtGradientForwardBackward] OrtModuleGraphBuilder.build")
        builder.build()

        if logger is not None:
            logger.info(
                "[OrtGradientForwardBackward] OrtModuleGraphBuilder.get_model")

        train_onnx_model_serialized = builder.get_model()

        optimized_pre_grad_model = builder.get_inference_optimized_model()
        graph_info = builder.get_graph_info()

        if logger is not None:
            logger.info("[OrtGradientForwardBackward] graph_info=%s",
                        OrtGradientForwardBackward._repr_helper_(
                            graph_info, indent=4))
            logger.info("[OrtGradientForwardBackward] create TrainSession")
            logger.info("[OrtGradientForwardBackward] sess_options=%s",
                        OrtGradientForwardBackward._repr_helper_(
                            self.sess_options, indent=4))
            logger.info(
                "[OrtGradientForwardBackward] providers=%r", self.providers)

        sess = InferenceSession(
            train_onnx_model_serialized, sess_options=self.sess_options,
            provider_options=self.provider_options, providers=self.providers)

        if logger is not None:
            logger.info("[OrtGradientForwardBackward] create InferenceSession")

        sess_eval = InferenceSession(
            optimized_pre_grad_model, sess_options=self.sess_options,
            provider_options=self.provider_options, providers=self.providers)

        if logger is not None:
            logger.info("[OrtGradientForwardBackward] create training agent")

        grad_input_names = [obj.name for obj in sess.get_inputs()]
        bw_fetches_names = [obj.name for obj in sess.get_outputs()]

        fw_outputs_device_info = [
            OrtDevice(
                OrtGradientForwardBackward._provider_name_to_device_type(i),
                OrtDevice.default_memory(), self.device_index)
            for i in self.providers]
        bw_outputs_device_info = [
            OrtDevice(
                OrtGradientForwardBackward._provider_name_to_device_type(
                    self.providers[0]),
                OrtDevice.default_memory(), self.device_index)
            for i in bw_fetches_names]
        fw_no_grad_output_device_info = [
            OrtDevice(
                OrtGradientForwardBackward._provider_name_to_device_type(
                    self.providers[0]),
                OrtDevice.default_memory(), self.device_index)
            for i in self.output_names]

        training_agent = TrainingAgent(
            sess._sess,
            grad_input_names,
            fw_outputs_device_info,
            bw_fetches_names,
            bw_outputs_device_info)

        if logger is not None:
            logger.info(
                "[OrtGradientForwardBackward] instantiate dynamic class %r",
                self.class_name)
            logger.info(
                "[OrtGradientForwardBackward] weights_to_train=%r",
                self.weights_to_train)
            logger.info(
                "[OrtGradientForwardBackward] grad_input_names=%r",
                grad_input_names)
            logger.info(
                "[OrtGradientForwardBackward] bw_fetches_names=%r",
                bw_fetches_names)
            logger.info(
                "[OrtGradientForwardBackward] device_index=%r",
                self.device_index)
        devices = list(fw_outputs_device_info)
        while len(devices) < len(grad_input_names):
            devices.append(devices[-1])

        trained_onnx = onnx.load(BytesIO(train_onnx_model_serialized))
        onnx_loss = onnx.load(BytesIO(optimized_pre_grad_model))
        for i, node in enumerate(trained_onnx.graph.node):
            if node.name == '':
                node.name = "N%d" % i
        for i, node in enumerate(onnx_loss.graph.node):
            if node.name == '':
                node.name = "N%d" % i

        kwargs = {
            '_run_options': self.run_options,
            '_sess': sess,
            '_sess_eval': sess_eval,
            '_training_agent': training_agent,
            '_cache': OrtValueCache(),
            '_logger': logger,
            '_input_names': self.input_names,
            '_grad_input_names': grad_input_names,
            '_output_names': self.output_names,
            '_bw_fetches_names': bw_fetches_names,
            '_fw_outputs_device_info': fw_outputs_device_info,
            '_bw_outputs_device_info': bw_outputs_device_info,
            '_fw_no_grad_output_device_info': fw_no_grad_output_device_info,
            '_weights_to_train': list(sorted(
                self.weights_to_train)),
            '_graph_info': graph_info,
            #
            '_trained_onnx': trained_onnx,
            '_optimized_pre_grad_model': onnx_loss,
            '_graph_builder': builder,
            '_devices': devices,
            '_debug': self.debug
        }
        graph = kwargs['_trained_onnx'].graph
        kwargs.update({
            '_onx_inp': [o.name for o in graph.input],
            '_onx_out': [o.name for o in graph.output]
        })

        if len(kwargs['_onx_inp']) != len(kwargs['_onx_out']):
            raise RuntimeError(  # pragma: no cover
                "Gradient input and output are inconsistant: "
                "%r != %r" % (kwargs['_onx_inp'], kwargs['_onx_out']))
        return kwargs


class OrtGradientForwardBackwardFunction:
    """
    Ancestor for a class implementing forward and backward
    and dynamically created by @see cl OrtGradientForwardBackward.

    Attributes stored in *forward* method:
    * `saved_tensors_`: list of tensors to save during forward
        and to retrieve during backward
    * `state_`: current weights stored in :epkg:`PartialGraphExecutionState`
    """

    def __init__(self):
        self.states_ = []
        self.saved_tensors_ = None

    @classmethod
    def save_onnx_graph(cls, folder, prefix=None, suffix=None):
        """
        Saves onnx graph stored in this class.
        """
        if prefix is None:
            prefix = ''  # pragma: no cover
        if suffix is None:
            suffix = ''  # pragma: no cover
        if isinstance(folder, str) and not os.path.exists(folder):
            raise FileNotFoundError(  # pragma: no cover
                "Folder %r does not exist." % folder)
        saved = {}
        for k, v in cls.__dict__.items():
            if hasattr(v, "SerializeToString"):
                if isinstance(folder, str):
                    name = "%s%s%s.%s.onnx" % (
                        prefix, cls.__name__, suffix, k)
                    filename = os.path.join(folder, name)
                    if os.path.exists(filename):
                        warnings.warn(  # pragma: no cover
                            "Filename %r already exists." % filename)
                    with open(filename, "wb") as f:
                        f.write(v.SerializeToString())
                    saved[k] = filename
                else:
                    saved[k] = v.SerializeToString()
            elif hasattr(v, "save_onnx_graph"):
                saved[k] = v.save_onnx_graph(
                    folder, prefix=prefix, suffix="%s.%s" % (suffix, k))
        return saved

    @staticmethod
    def device_name(device):
        """
        Returns the device name of a device.

        :param device: OrtDevice
        :return: string
        """
        if device.device_type() == OrtDevice.cpu():
            return 'Cpu'
        if device.device_type() == OrtDevice.cuda():  # pragma: no cover
            return 'Gpu'
        raise RuntimeError(  # pragma: no cover
            "Unexpected value for device type %r." % device.device_type())

    @staticmethod
    def input_to_ort(tensors, devices, debug):
        "Converts a list of tensos into an :epkg:`OrtValueVector`."
        def _validate_(tensors):
            if any(map(
                    lambda tu: (
                        tu[0].device_name() !=
                        OrtGradientForwardBackwardFunction.device_name(
                            tu[1])),
                    zip(tensors, devices))):
                raise RuntimeError(  # pragma: no cover
                    "Not all inputs are on the same device %r != %r." % (
                        [OrtGradientForwardBackward.device_name(d)
                         for d in devices],
                        [x.device_name() for x in tensors]))

        if isinstance(tensors, OrtValueVector):
            if debug:
                _validate_(tensors)
            return tensors
        if all(map(lambda t: isinstance(t, C_OrtValue), tensors)):
            if debug:
                _validate_(tensors)
            vect = OrtValueVector()
            vect.reserve(len(tensors))
            for t in tensors:
                if t is None:
                    raise NotImplementedError(  # pragma: no cover
                        "Empty vector found.")
                vect.push_back(t)
            return vect

        # generic case
        vect = OrtValueVector()
        vect.reserve(len(tensors))
        for t, dev in zip(tensors, devices):
            if t is None:
                # if gradient then
                # grad_output = torch.zeros(shape, device=device, dtype=dtype)
                raise NotImplementedError(  # pragma: no cover
                    "Empty vector found.")
            if not t.data.contiguous:
                t = t.as_contiguous()  # pragma: no cover
            vect.push_back(C_OrtValue.ortvalue_from_numpy(t, dev))
        if debug:
            if len(vect) != len(tensors):
                raise RuntimeError(  # pragma: no cover
                    "Unexpected array length %d != %d (len(devices)=%d)." % (
                        len(vect), len(tensors), len(devices)))
            _validate_(vect)
        return vect

    def save_for_backward(self, inputs):
        """
        Saves inputs furing forward steps. The list inputs
        is copied (simple copy, no deep copy).

        :param inputs: list of tensors to save.
        """
        self.saved_tensors_ = list(inputs)

    @property
    def saved_tensors(self):
        """
        Returns saved tensors during forward step.
        """
        if self.saved_tensors_ is None:
            raise RuntimeError(  # pragma: no cover
                "No tensors was saved with save_for_backward.")
        return self.saved_tensors_

    def forward(self, inputs, training=False, forward_outputs_cache=None):
        """
        Implements forward function.

        :param inputs: inputs
        :param training: only inference or training as well
        :return: output as :epkg:`OrtValueVector`
        """
        logger = self._logger
        cls = self.__class__

        def _log(msg, *args):
            logger.debug("[%s.forward] (%dI) " + msg,
                         cls.__name__, len(inputs), *args)

        if logger is not None:
            if training:
                _log("begin with gradient")
            else:
                _log("begin")
            _log("torch function %r", type(cls))
            _log("ort class %r", cls)
            _log("create OrtValueVector (through dlpack)")

        forward_inputs = cls.input_to_ort(
            inputs, cls._devices, cls._debug)

        if training:
            forward_outputs = forward_outputs_cache or OrtValueVector()
            state = PartialGraphExecutionState()
            self.states_.append(state)
            if logger is not None:
                _log("run_forward")
            cls._training_agent.run_forward(
                forward_inputs, forward_outputs, state, cls._cache)

            self.save_for_backward(inputs)
            if logger is not None:
                _log("end")
            return forward_outputs
        else:
            # what about bind_input (+ data_ptr)
            if len(forward_inputs) != len(cls._grad_input_names):
                raise RuntimeError(  # pragma: no cover
                    "Size mismatch len(inputs)=%d, len(onnx inputs)=%d." % (
                        len(forward_inputs), len(cls._grad_input_names)))
            iobinding = SessionIOBinding(cls._sess_eval._sess)
            if logger is not None:
                _log("bind inputs %r", cls._grad_input_names)
            for name, inp in zip(
                    cls._grad_input_names, forward_inputs):
                iobinding.bind_ortvalue_input(name, inp)

            # bind output
            if logger is not None:
                _log("bind outputs %r", cls._output_names)
            for name, dev in zip(
                    cls._output_names, cls._fw_no_grad_output_device_info):
                iobinding.bind_output(name, dev)

            # if the shape is known in advance
            # iobinding.bind_output(
            #    output_desc.name, torch_tensor.device.type,
            #    _utils.get_device_index(target_device),
            #    _utils.dtype_torch_to_numpy(torch_tensor.dtype),
            #    list(torch_tensor.size()), torch_tensor.data_ptr())

            if logger is not None:
                _log("grad_enabled=False (run_with_iobinding)")
            cls._sess_eval._sess.run_with_iobinding(
                iobinding, cls._run_options)
            if logger is not None:
                _log("get_outputs")
            ortvalues = iobinding.get_outputs()
            if logger is not None:
                _log("to torck.tensor (%d)", len(ortvalues))
                _log("end")
            return ortvalues

    def backward(self, grad_outputs, backward_outputs_cache=None):
        """
        Implements backward function. The function returns
        an :epkg:`OrtValueVector`.
        """
        cls = self.__class__
        logger = cls._logger

        def _log(msg, *args):
            logger.debug("[%s.backward] (%dI) " + msg,
                         cls.__name__, len(grad_outputs), *args)

        if logger is not None:
            _log("begin")
            _log("torch function %r", type(cls))
            _log("ort class %r", cls)
            _log("saved_tensors")

        inputs = self.saved_tensors
        if logger is not None:
            _log("DEBUG: saved_tensors %r", type(inputs))
            _log("self.state_.pop()")
        state = self.states_.pop()

        if logger is not None:
            _log("create OrtValueVector")

        backward_inputs = cls.input_to_ort(
            grad_outputs, cls._bw_outputs_device_info, cls._debug)

        if logger is not None:
            _log("len(grad_outputs)=%d type(grad_outputs)=%r",
                 len(grad_outputs), type(grad_outputs))
            _log("len(backward_inputs)=%d type(backward_inputs)=%r",
                 len(backward_inputs), type(backward_inputs))
            for i in range(len(backward_inputs)):  # pylint: disable=C0200
                _log("backward_inputs[%d].shape=%r",
                     i, backward_inputs[i].shape())
            _log("run_backward")
        backward_outputs = backward_outputs_cache or OrtValueVector()
        cls._training_agent.run_backward(
            backward_inputs, backward_outputs, state)
        if logger is not None:  # pragma: no cover
            _log("DEBUG")
            for i, ov in enumerate(backward_outputs):
                _log("BCK-RET: i=%d - shape=%r - ptr=%r",
                     i, ov.shape(), ov.data_ptr())
            _log("got %r gradients", len(backward_outputs))
            _log("end")
        return backward_outputs



# pylint: disable=C0415,E0611
"""
@file
@brief Onnx helper.
"""
import numpy
from .onnx_helper import dtype_to_var_type, add_initializer


def get_supported_functions():
    """
    Returns the list of supported function by @see fn function_onnx_graph.
    """
    glo = globals()
    res = {}
    for k, v in glo.items():
        if k.startswith('_onnx_'):
            res[k[6:]] = v.__doc__
    return res


def function_onnx_graph(name, target_opset=None, dtype=numpy.float32,
                        weight_name=None, **kwargs):
    """
    Returns the ONNX graph corresponding to a function.

    :param name: name
    :param target_opset: opset version
    :param dtype: computation type
    :param weight_name: weight name if any
    :param kwargs: additional parameters
    :return: ONNX graph

    A wrong name will raise an exception giving the whole of
    supported function. One example with function `square_error`:

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('square_error')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())

    An example on how to use it:

    .. runpython::
        :showcode:

        import numpy
        from onnxruntime import InferenceSession
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('square_error')
        sess = InferenceSession(model_onnx.SerializeToString())
        res = sess.run(None, {
            'X1': numpy.array([[0, 1]], dtype=numpy.float32).T,
            'X2': numpy.array([[1, 2]], dtype=numpy.float32).T})
        print(res[0])

    List of supported functions:

    .. runpython::
        :showcode:

        from onnxcustom.utils.onnx_function import get_supported_functions
        print("\\n".join(sorted(get_supported_functions())))
    """
    glo = globals()
    full_name = "_onnx_" + name
    if full_name in glo:
        if weight_name is None:
            return glo[full_name](target_opset=target_opset,
                                  dtype=dtype, **kwargs)
        return glo[full_name](target_opset=target_opset, dtype=dtype,
                              weight_name=weight_name, **kwargs)
    raise ValueError(
        "Unable to find function %r in %r." % (
            full_name, list(sorted(
                k for k in glo if k.startswith('_onnx_')))))


def _onnx_axpy(target_opset=None, dtype=numpy.float32):
    """
    Returns the ONNX graph for function
    :math:`Y = f(X1, X2, \\alpha) = \\alpha X1 + X2`.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('axpy')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import OnnxAdd, OnnxMul
    res = OnnxAdd(OnnxMul('X1', 'alpha', op_version=target_opset),
                  'X2', op_version=target_opset, output_names=['Y'])
    var_type = dtype_to_var_type(dtype)
    varsx = [('X1', var_type()), ('X2', var_type()),
             ('alpha', var_type([1]))]
    onx = res.to_onnx(varsx, outputs=[('Y', var_type())],
                      target_opset=target_opset)
    return onx


def _onnx_axpyw(target_opset=None, dtype=numpy.float32):
    """
    Returns the ONNX graph for function
    :math:`Y, Z = f(X1, X2, G, \\alpha, \\beta) = (Y, Z)`
    where :math:`Z = \\beta G + \\alpha X1` and
    :math:`Y = Z + X2`.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('axpy')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import OnnxAdd, OnnxMul
    s1 = OnnxMul('X1', 'alpha', op_version=target_opset)
    s2 = OnnxMul('G', 'beta', op_version=target_opset)
    Z = OnnxAdd(s1, s2, op_version=target_opset,
                output_names=['Z'])
    Y = OnnxAdd(Z, 'X2', op_version=target_opset, output_names=['Y'])
    var_type = dtype_to_var_type(dtype)
    varsx = [('X1', var_type()), ('X2', var_type()),
             ('G', var_type()),
             ('alpha', var_type([1])), ('beta', var_type([1]))]
    onx = Y.to_onnx(
        varsx, outputs=[('Y', var_type()), ('Z', var_type())],
        target_opset=target_opset, other_outputs=[Z])
    return onx


def _onnx_axpyw2(target_opset=None, dtype=numpy.float32):
    """
    Returns the ONNX graph for function
    :math:`Y, Z = f(X1, X2, G, \\alpha, \\beta) = (Y, Z)`
    where :math:`Z = \\beta G + \\alpha X1` and
    :math:`Y = \\beta * Z + \\alpha X1 + X2`.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('axpy')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import OnnxAdd, OnnxMul
    s1 = OnnxMul('X1', 'alpha', op_version=target_opset)
    s2 = OnnxMul('G', 'beta', op_version=target_opset)
    Z = OnnxAdd(s1, s2, op_version=target_opset,
                output_names=['Z'])
    s2_2 = OnnxMul(Z, 'beta', op_version=target_opset)
    s2_3 = OnnxAdd(s1, s2_2, op_version=target_opset)
    Y = OnnxAdd(s2_3, 'X2', op_version=target_opset, output_names=['Y'])
    var_type = dtype_to_var_type(dtype)
    varsx = [('X1', var_type()), ('X2', var_type()),
             ('G', var_type()),
             ('alpha', var_type([1])), ('beta', var_type([1]))]
    onx = Y.to_onnx(
        varsx, outputs=[('Y', var_type()), ('Z', var_type())],
        target_opset=target_opset, other_outputs=[Z])
    return onx


def _onnx_square_error(target_opset=None, dtype=numpy.float32,
                       weight_name=None):
    """
    Returns the ONNX graph for function
    :math:`Y = f(X1, X2) = \\lVert X1 - X2 \\rVert ^2` or
    :math:`Y = f(X1, X2) = \\lVert X1 - X2 \\rVert ^2 w` if
    *weight_name* is not None

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('square_error')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import (
        OnnxSub, OnnxReduceSumSquare, OnnxReshape,
        OnnxReduceSum, OnnxMul)
    diff = OnnxSub('X1', 'X2', op_version=target_opset)
    if weight_name is None:
        res = OnnxReduceSumSquare(diff, op_version=target_opset)
    else:
        mul = OnnxMul(
            OnnxMul(diff, diff, op_version=target_opset),
            OnnxReshape(weight_name,
                        numpy.array([-1, 1], dtype=numpy.int64),
                        op_version=target_opset),
            op_version=target_opset)
        res = OnnxReduceSum(mul, op_version=target_opset)
    res = OnnxReshape(res, numpy.array([-1], numpy.int64),
                      op_version=target_opset,
                      output_names=['Y'])
    var_type = dtype_to_var_type(dtype)
    varsx = [('X1', var_type([None, None])),
             ('X2', var_type([None, None]))]
    if weight_name is not None:
        varsx.append((weight_name, var_type([None])))
    onx = res.to_onnx(varsx, outputs=[('Y', var_type())],
                      target_opset=target_opset)
    if weight_name is not None:
        onx = add_initializer(
            onx, weight_name, numpy.array([1], dtype=dtype))
    return onx


def _onnx_grad_square_error(target_opset=None, dtype=numpy.float32,
                            weight_name=None):
    """
    Returns the ONNX graph for the gradient of function
    :math:`Y = f(X1, X2) = \\lVert X1 - X2 \\rVert ^2` or
    :math:`Y = f(X1, X2) = \\lVert X1 - X2 \\rVert ^2 w` if
    *weight_name* is not None

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('grad_square_error')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import OnnxSub, OnnxMul, OnnxReshape
    diff = OnnxSub('X1', 'X2', op_version=target_opset)
    if weight_name is None:
        res = OnnxMul(diff, numpy.array([-2], dtype=dtype),
                      op_version=target_opset, output_names=['Y'])
    else:
        res = OnnxMul(
            OnnxMul(diff, numpy.array([-2], dtype=dtype),
                    op_version=target_opset),
            OnnxReshape(weight_name,
                        numpy.array([-1, 1], dtype=numpy.int64),
                        op_version=target_opset),
            op_version=target_opset, output_names=['Y'])
    var_type = dtype_to_var_type(dtype)
    varsx = [('X1', var_type([None, None])), ('X2', var_type([None, None]))]
    if weight_name is not None:
        varsx.append((weight_name, var_type([None])))
    onx = res.to_onnx(varsx, outputs=[('Y', var_type())],
                      target_opset=target_opset)
    if weight_name is not None:
        onx = add_initializer(
            onx, weight_name, numpy.array([1], dtype=dtype))
    return onx


def _onnx_copy(target_opset=None, dtype=numpy.float32):
    """
    Returns the ONNX graph for function
    :math:`Y = X`.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('copy')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import OnnxIdentity
    res = OnnxIdentity('X', op_version=target_opset,
                       output_names=['Y'])
    var_type = dtype_to_var_type(dtype)
    varsx = [('X', var_type())]
    onx = res.to_onnx(varsx, outputs=[('Y', var_type())],
                      target_opset=target_opset)
    return onx


def _onnx_zero(target_opset=None, dtype=numpy.float32):
    """
    Returns the ONNX graph for function
    :math:`Y = X * 0`.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('zero')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import OnnxMul
    res = OnnxMul('X', numpy.array([0], dtype=dtype),
                  op_version=target_opset,
                  output_names=['Y'])
    var_type = dtype_to_var_type(dtype)
    varsx = [('X', var_type())]
    onx = res.to_onnx(varsx, outputs=[('Y', var_type())],
                      target_opset=target_opset)
    return onx


def _onnx_linear_regression(target_opset=None, dtype=numpy.float32):
    """
    Returns the ONNX graph for function
    :math:`Y = f(X, A, B) = A X + B`.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('linear_regression')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import (
        OnnxMatMul, OnnxAdd)
    res = OnnxAdd(
        OnnxMatMul('X', 'A', op_version=target_opset),
        'B', op_version=target_opset, output_names=['Y'])

    var_type = dtype_to_var_type(dtype)
    varsx = [('X', var_type([None, None])),
             ('A', var_type([None, None])),
             ('B', var_type([None, None]))]
    onx = res.to_onnx(
        varsx, outputs=[('Y', var_type())],
        target_opset=target_opset, other_outputs=[res])
    return onx


def _onnx_grad_loss_square_error(target_opset=None, dtype=numpy.float32,
                                 weight_name=None, multiply=2):
    """
    Returns the ONNX graph for function
    :math:`Y = f(X1, X2) = \\lVert (X1 - X2) \\rVert ^2` or
    :math:`Y = f(X1, X2) = \\lVert (\\sqrt{w}(X1 - X2) \\rVert ^2 w` if
    *weight_name* is not None and its gradient.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('grad_loss_square_error')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import (
        OnnxSub, OnnxReduceSumSquare, OnnxMul,
        OnnxReduceSum, OnnxReshape)
    diff = OnnxSub('X1', 'X2', op_version=target_opset)
    if weight_name is None:
        res = OnnxMul(OnnxReduceSumSquare(diff, op_version=target_opset),
                      numpy.array([multiply * 0.5], dtype=numpy.float32),
                      op_version=target_opset)
        res2 = OnnxMul(diff, numpy.array([-multiply], dtype=dtype),
                       op_version=target_opset, output_names=['Z'])
    else:
        resh = OnnxReshape(weight_name,
                           numpy.array([-1, 1], dtype=numpy.int64),
                           op_version=target_opset)
        mul = OnnxMul(
            OnnxMul(diff, diff, op_version=target_opset),
            resh, op_version=target_opset)
        res = OnnxMul(OnnxReduceSum(mul, op_version=target_opset),
                      numpy.array([multiply * 0.5], dtype=numpy.float32),
                      op_version=target_opset)

        res2 = OnnxMul(
            OnnxMul(diff, numpy.array([-multiply], dtype=dtype),
                    op_version=target_opset),
            resh, op_version=target_opset, output_names=['Z'])

    res = OnnxReshape(res, numpy.array([-1], numpy.int64),
                      op_version=target_opset,
                      output_names=['Y'])

    var_type = dtype_to_var_type(dtype)
    varsx = [('X1', var_type([None, None])),
             ('X2', var_type([None, None]))]
    if weight_name is not None:
        varsx.append((weight_name, var_type([None])))
    onx = res.to_onnx(
        varsx, outputs=[('Y', var_type()), ('Z', var_type())],
        target_opset=target_opset, other_outputs=[res2])
    if weight_name is not None:
        onx = add_initializer(
            onx, weight_name, numpy.array([1], dtype=dtype))
    return onx


def _onnx_grad_loss_absolute_error(target_opset=None, dtype=numpy.float32,
                                   weight_name=None):
    """
    Returns the ONNX graph for function
    :math:`Y = f(X1, X2) = \\lVert X1 - X2 \\rVert` or
    :math:`Y = f(X1, X2) = \\lVert (X1 - X2)w \\rVert` if
    *weight_name* is not None and its gradient.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('grad_loss_absolute_error')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import (
        OnnxSub, OnnxMul, OnnxReduceSum, OnnxReshape,
        OnnxSign, OnnxAbs)
    diff = OnnxSub('X1', 'X2', op_version=target_opset)
    abs_diff = OnnxAbs(diff, op_version=target_opset)
    if weight_name is None:
        res = OnnxReduceSum(abs_diff, op_version=target_opset)
        res2 = OnnxSign(diff, op_version=target_opset,
                        output_names=['Z'])
    else:
        resh = OnnxReshape(weight_name,
                           numpy.array([-1, 1], dtype=numpy.int64),
                           op_version=target_opset)
        mul = OnnxMul(abs_diff, resh, op_version=target_opset)
        res = OnnxReduceSum(mul, op_version=target_opset)
        res2 = OnnxMul(
            OnnxSign(diff, op_version=target_opset),
            resh, op_version=target_opset, output_names=['Z'])

    res = OnnxReshape(res, numpy.array([-1], numpy.int64),
                      op_version=target_opset,
                      output_names=['Y'])
    var_type = dtype_to_var_type(dtype)
    varsx = [('X1', var_type([None, None])),
             ('X2', var_type([None, None]))]
    if weight_name is not None:
        varsx.append((weight_name, var_type([None])))
    onx = res.to_onnx(
        varsx, outputs=[('Y', var_type()), ('Z', var_type())],
        target_opset=target_opset, other_outputs=[res2])
    if weight_name is not None:
        onx = add_initializer(
            onx, weight_name, numpy.array([1], dtype=dtype))
    return onx


def _onnx_grad_loss_elastic_error(target_opset=None, dtype=numpy.float32,
                                  weight_name=None,
                                  l1_weight=0.01, l2_weight=0.01):
    """
    Returns the ONNX graph for function
    :math:`Y = f(X1, X2) = \\beta \\lVert X1 - X2 \\rVert +
    \\alpha \\lVert X1 - X2 \\rVert^2` or
    :math:`Y = f(X1, X2) = \\beta \\lVert w(X1 - X2) \\rVert +
    \\alpha \\lVert (\\sqrt{w})(X1 - X2) \\rVert^2` if
    *weight_name* is not None and its gradient.
    *l1_weight* is :math:`\\beta` and
    *l2_weight* is :math:`\\alpha`.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('grad_loss_elastic_error')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import (
        OnnxSub, OnnxMul, OnnxAdd, OnnxReduceSumSquare,
        OnnxReduceSum, OnnxReshape, OnnxSign, OnnxAbs)
    diff = OnnxSub('X1', 'X2', op_version=target_opset)
    abs_diff = OnnxAbs(diff, op_version=target_opset)
    if weight_name is None:
        res_l1 = OnnxReduceSum(abs_diff, op_version=target_opset)
        res2_l1 = OnnxSign(diff, op_version=target_opset)
        res_l2 = OnnxReduceSumSquare(diff, op_version=target_opset)
        res2_l2 = diff
    else:
        resh = OnnxReshape(weight_name,
                           numpy.array([-1, 1], dtype=numpy.int64),
                           op_version=target_opset)
        mul = OnnxMul(abs_diff, resh, op_version=target_opset)
        res_l1 = OnnxReduceSum(mul, op_version=target_opset)
        res2_l1 = OnnxMul(
            OnnxSign(diff, op_version=target_opset),
            resh, op_version=target_opset)

        mul = OnnxMul(
            OnnxMul(diff, diff, op_version=target_opset),
            resh, op_version=target_opset)
        res_l2 = OnnxReduceSum(mul, op_version=target_opset)
        res2_l2 = OnnxMul(diff, resh, op_version=target_opset)

    res = OnnxAdd(
        OnnxMul(res_l1, numpy.array([l1_weight], dtype=dtype),
                op_version=target_opset),
        OnnxMul(res_l2, numpy.array([l2_weight], dtype=dtype),
                op_version=target_opset),
        op_version=target_opset)
    res = OnnxReshape(res, numpy.array([-1], numpy.int64),
                      op_version=target_opset,
                      output_names=['Y'])

    res2 = OnnxAdd(
        OnnxMul(res2_l1, numpy.array([l1_weight], dtype=dtype),
                op_version=target_opset),
        OnnxMul(res2_l2, numpy.array([l2_weight * (-2)], dtype=dtype),
                op_version=target_opset),
        op_version=target_opset, output_names=['Z'])

    var_type = dtype_to_var_type(dtype)
    varsx = [('X1', var_type([None, None])),
             ('X2', var_type([None, None]))]
    if weight_name is not None:
        varsx.append((weight_name, var_type([None])))
    onx = res.to_onnx(
        varsx, outputs=[('Y', var_type()), ('Z', var_type())],
        target_opset=target_opset, other_outputs=[res2])
    if weight_name is not None:
        onx = add_initializer(
            onx, weight_name, numpy.array([1], dtype=dtype))
    return onx


def _onnx_grad_penalty_elastic_error(target_opset=None, dtype=numpy.float32,
                                     l1_weight=0.01, l2_weight=0.01):
    """
    Returns the ONNX graph for function
    :math:`Y = f(W) = \\beta \\lVert W \\rVert +
    \\alpha \\lVert W \\rVert^2`
    *l1_weight* is :math:`\\beta` and
    *l2_weight* is :math:`\\alpha`.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('grad_penalty_elastic_error')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import (
        OnnxMul, OnnxAdd, OnnxReduceSumSquare,
        OnnxReduceSum, OnnxSign, OnnxAbs, OnnxReshape)
    diff = 'X'
    abs_diff = OnnxAbs(diff, op_version=target_opset)
    res_l1 = OnnxReduceSum(abs_diff, op_version=target_opset)
    res2_l1 = OnnxSign(diff, op_version=target_opset)
    res_l2 = OnnxReduceSumSquare(diff, op_version=target_opset)
    res2_l2 = diff

    res = OnnxAdd(
        OnnxMul(res_l1, numpy.array([l1_weight], dtype=dtype),
                op_version=target_opset),
        OnnxMul(res_l2, numpy.array([l2_weight], dtype=dtype),
                op_version=target_opset),
        op_version=target_opset)
    res = OnnxReshape(res, numpy.array([-1], numpy.int64),
                      op_version=target_opset,
                      output_names=['Y'])

    res2 = OnnxAdd(
        OnnxMul(res2_l1, numpy.array([l1_weight], dtype=dtype),
                op_version=target_opset),
        OnnxMul(res2_l2, numpy.array([l2_weight * (2)], dtype=dtype),
                op_version=target_opset),
        op_version=target_opset, output_names=['Z'])

    var_type = dtype_to_var_type(dtype)
    varsx = [('X', var_type([None, None]))]
    onx = res.to_onnx(
        varsx, outputs=[('Y', var_type([None])), ('Z', var_type())],
        target_opset=target_opset, other_outputs=[res2])
    return onx


def _onnx_n_penalty_elastic_error(target_opset=None, dtype=numpy.float32,
                                  weight_name=None,
                                  l1_weight=0.01, l2_weight=0.01, n_tensors=1,
                                  loss_shape=(1, 1)):
    """
    Returns the ONNX graph for function
    :math:`Y = f(W) = \\beta \\lVert W \\rVert +
    \\alpha \\lVert W \\rVert^2`
    *l1_weight* is :math:`\\beta` and
    *l2_weight* is :math:`\\alpha`.
    It does that for *n_tensors* and adds all of the results
    to an input loss.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph(
            'n_penalty_elastic_error', n_tensors=2)
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import (
        OnnxMul, OnnxAdd, OnnxReduceSumSquare,
        OnnxReduceSum, OnnxAbs, OnnxReshape)

    if n_tensors <= 0:
        raise ValueError(  # pragma: no cover
            "This function is useless if the number of tensors is null.")

    var_type = dtype_to_var_type(dtype)
    varsx = [('loss', var_type(loss_shape))]
    names = ['loss']
    for n in range(n_tensors):
        name = 'W%d' % n
        abs_diff = OnnxAbs(name, op_version=target_opset)
        res_l1 = OnnxReduceSum(abs_diff, op_version=target_opset)
        # res2_l1 = OnnxSign(diff, op_version=target_opset)
        res_l2 = OnnxReduceSumSquare(name, op_version=target_opset)
        # res2_l2 = diff
        res = OnnxAdd(
            OnnxMul(res_l1, numpy.array([l1_weight], dtype=dtype),
                    op_version=target_opset),
            OnnxMul(res_l2, numpy.array([l2_weight], dtype=dtype),
                    op_version=target_opset),
            op_version=target_opset)
        names.append(res)
        varsx.append(('W%d' % n, var_type()))

    if len(names) == 2:
        res = OnnxAdd(*names, op_version=target_opset)
    else:
        res = OnnxAdd(names[1], names[2], op_version=target_opset)
        for i in range(3, len(names)):
            res = OnnxAdd(res, names[i], op_version=target_opset)
        res = OnnxAdd(names[0], res, op_version=target_opset)

    res = OnnxReshape(res, numpy.array([-1], numpy.int64),
                      op_version=target_opset,
                      output_names=['Y'])
    onx = res.to_onnx(
        varsx, outputs=[('Y', var_type([None]))],
        target_opset=target_opset)
    return onx


def _onnx_update_penalty_elastic_error(target_opset=None, dtype=numpy.float32,
                                       l1=1e-4, l2=1e-4):
    """
    Returns the ONNX graph for function
    :math:`Y = f(W) = W - 2 \\beta W - \\alpha sign(W)`
    *l1* is :math:`\\beta` and
    *l2* is :math:`\\alpha`.

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph(
            'update_penalty_elastic_error')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from skl2onnx.algebra.onnx_ops import (
        OnnxSub, OnnxMul, OnnxSign)

    res = OnnxSub(
        OnnxMul('X', numpy.array([1 - 2 * l2], dtype=dtype),
                op_version=target_opset),
        OnnxMul(OnnxSign('X', op_version=target_opset),
                numpy.array([l1], dtype=dtype),
                op_version=target_opset),
        op_version=target_opset,
        output_names=['Y'])

    var_type = dtype_to_var_type(dtype)
    varsx = [('X', var_type())]
    onx = res.to_onnx(
        varsx, outputs=[('Y', var_type())],
        target_opset=target_opset)
    return onx


def _onnx_grad_sigmoid_neg_log_loss_error(target_opset=None,
                                          dtype=numpy.float32,
                                          eps=1e-5,
                                          weight_name=None):
    """
    The function the raw scores from a classifier, uses the
    sigmoid function to compute probabilities, then the log function
    to compute the loss. It creates the ONNX graph for this function
    and the associated gradient of the loss against the raw scores.

    Probabilites (class 1): :math:`p(s) = \\frac{1}{1 + \\exp(-s)}`.
    Loss (for two classes): :math:`L(y, s) = (1 - y)\\log(1 - p(s)) +
    y \\log(p(s))`.
    Gradient :math:`\\frac{dL(y, s)}{ds} = y - p(s)`.
    To avoid nan values, probabilies are clipped:
    :math:`p(s) = \\max(\\min(p(s), 1 - \\epsilon), \\epsilon)`.
    :math:`y \\in \\{0, 1\\}` (integer). *s* is a float.

    :param eps: to clip probabilities and avoid computing `log(0)`

    .. gdot::
        :script: DOT-SECTION

        from mlprodict.onnxrt import OnnxInference
        from onnxcustom.utils.onnx_function import function_onnx_graph

        model_onnx = function_onnx_graph('grad_sigmoid_neg_log_loss_error')
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE
    from skl2onnx.algebra.onnx_ops import (
        OnnxSub, OnnxMul, OnnxSigmoid, OnnxLog, OnnxNeg,
        OnnxReduceMean, OnnxReshape, OnnxAdd, OnnxCast, OnnxClip)

    p1c = OnnxSigmoid('X2', op_version=target_opset)
    p1 = OnnxClip(p1c, numpy.array([eps], dtype=dtype),
                  numpy.array([1 - eps], dtype=dtype),
                  op_version=target_opset)
    p0 = OnnxSub(numpy.array([1], dtype=dtype), p1,
                 op_version=target_opset)
    y1 = OnnxCast('X1', to=NP_TYPE_TO_TENSOR_TYPE[numpy.dtype(dtype)],
                  op_version=target_opset)
    y0 = OnnxSub(numpy.array([1], dtype=dtype), y1,
                 op_version=target_opset)
    loss_obs = OnnxAdd(
        OnnxMul(y0, OnnxLog(p0, op_version=target_opset),
                op_version=target_opset),
        OnnxMul(y1, OnnxLog(p1, op_version=target_opset),
                op_version=target_opset),
        op_version=target_opset)

    if weight_name is None:
        loss = OnnxReduceMean(loss_obs, op_version=target_opset)
        grad = OnnxSub(p1, y1, op_version=target_opset,
                       output_names=['Z'])
    else:
        loss = OnnxReduceMean(
            OnnxMul(loss_obs, weight_name, op_version=target_opset),
            op_version=target_opset)
        grad = OnnxMul(
            OnnxSub(p1, y1, op_version=target_opset),
            OnnxReshape(weight_name, numpy.array([-1, 1], dtype=numpy.int64),
                        op_version=target_opset),
            output_names=['Z'], op_version=target_opset)

    loss_neg = OnnxNeg(loss, op_version=target_opset)
    res = OnnxReshape(loss_neg, numpy.array([-1], numpy.int64),
                      op_version=target_opset,
                      output_names=['Y'])

    var_type_int64 = dtype_to_var_type(numpy.int64)
    var_type = dtype_to_var_type(dtype)
    varsx = [('X1', var_type_int64([None, None])),
             ('X2', var_type([None, None]))]
    if weight_name is not None:
        varsx.append((weight_name, var_type([None])))
    onx = res.to_onnx(
        varsx, outputs=[('Y', var_type()), ('Z', var_type())],
        target_opset=target_opset, other_outputs=[grad])
    if weight_name is not None:
        onx = add_initializer(
            onx, weight_name, numpy.array([1], dtype=dtype))
    return onx



"""
@file
@brief Manipulate data for training.
"""
import numpy
from ..utils.onnxruntime_helper import (
    get_ort_device, numpy_to_ort_value, ort_device_to_string)


class OrtDataLoader:
    """
    Draws consecutive random observations from a dataset
    by batch. It iterates over the datasets by drawing
    *batch_size* consecutive observations.

    :param X: features
    :param y: labels
    :param sample_weight: weight or None
    :param batch_size: batch size (consecutive observations)
    :param device: :epkg:`C_OrtDevice` or a string such as `'cpu'`
    :param random_iter: random iteration

    See example :ref:`l-orttraining-nn-gpu`.
    """

    def __init__(self, X, y, sample_weight=None,
                 batch_size=20, device='cpu', random_iter=True):
        if len(y.shape) == 1:
            y = y.reshape((-1, 1))
        if X.shape[0] != y.shape[0]:
            raise ValueError(  # pragma: no cover
                "Shape mismatch X.shape=%r, y.shape=%r." % (X.shape, y.shape))

        self.batch_size = batch_size
        self.device = get_ort_device(device)
        self.random_iter = random_iter

        self.X_np = numpy.ascontiguousarray(X)
        self.y_np = numpy.ascontiguousarray(y).reshape((-1, 1))

        self.X_ort = numpy_to_ort_value(self.X_np, self.device)
        self.y_ort = numpy_to_ort_value(self.y_np, self.device)

        self.desc = [(self.X_np.shape, self.X_np.dtype),
                     (self.y_np.shape, self.y_np.dtype)]

        if sample_weight is None:
            self.w_np = None
            self.w_ort = None
        else:
            if X.shape[0] != sample_weight.shape[0]:
                raise ValueError(  # pragma: no cover
                    "Shape mismatch X.shape=%r, sample_weight.shape=%r."
                    "" % (X.shape, sample_weight.shape))
            self.w_np = numpy.ascontiguousarray(
                sample_weight).reshape((-1, ))
            self.w_ort = numpy_to_ort_value(self.w_np, self.device)
            self.desc.append((self.w_np.shape, self.w_np.dtype))

    def __getstate__(self):
        "Removes any non pickable attribute."
        state = {}
        for att in ['X_np', 'y_np', 'w_np',
                    'desc', 'batch_size', 'random_iter']:
            state[att] = getattr(self, att)
        state['device'] = ort_device_to_string(self.device)
        return state

    def __setstate__(self, state):
        "Restores any non pickable attribute."
        for att, v in state.items():
            setattr(self, att, v)
        self.device = get_ort_device(self.device)
        self.X_ort = numpy_to_ort_value(self.X_np, self.device)
        self.y_ort = numpy_to_ort_value(self.y_np, self.device)
        if self.w_np is None:
            self.w_ort = None
        else:
            self.w_ort = numpy_to_ort_value(
                self.w_np, self.device)
        return self

    def __repr__(self):
        "usual"
        return "%s(..., ..., batch_size=%r, device=%r)" % (
            self.__class__.__name__, self.batch_size,
            ort_device_to_string(self.device))

    def __len__(self):
        "Returns the number of observations."
        return self.desc[0][0][0]

    def _next_iter(self, previous):
        if self.random_iter:
            b = len(self) - self.batch_size
            return numpy.random.randint(0, b)
        if previous == -1:
            return 0
        i = previous + self.batch_size
        if i + self.batch_size > len(self):
            i = len(self) - self.batch_size
        return i

    def iter_numpy(self):
        """
        Iterates over the datasets by drawing
        *batch_size* consecutive observations.
        This iterator is slow as it copies the data of every
        batch. The function yields :epkg:`C_OrtValue`.
        """
        if self.device.device_type() != self.device.cpu():
            raise RuntimeError(  # pragma: no cover
                "Only CPU device is allowed if numpy arrays are requested "
                "not %r." % ort_device_to_string(self.device))
        N = 0
        b = len(self) - self.batch_size
        if self.w_np is None:
            if b <= 0 or self.batch_size <= 0:
                yield (self.X_np, self.y_np)
            else:
                i = -1
                while N < len(self):
                    i = self._next_iter(i)
                    N += self.batch_size
                    yield (self.X_np[i:i + self.batch_size],
                           self.y_np[i:i + self.batch_size])
        else:
            if b <= 0 or self.batch_size <= 0:
                yield (self.X_np, self.y_np, self.w_np)
            else:
                i = -1
                while N < len(self):
                    i = self._next_iter(i)
                    N += self.batch_size
                    yield (self.X_np[i:i + self.batch_size],
                           self.y_np[i:i + self.batch_size],
                           self.w_np[i:i + self.batch_size])

    def iter_ortvalue(self):
        """
        Iterates over the datasets by drawing
        *batch_size* consecutive observations.
        This iterator is slow as it copies the data of every
        batch. The function yields :epkg:`C_OrtValue`.
        """
        N = 0
        b = len(self) - self.batch_size
        if self.w_ort is None:
            if b <= 0 or self.batch_size <= 0:
                yield (self.X_ort, self.y_ort)
            else:
                i = -1
                while N < len(self):
                    i = self._next_iter(i)
                    N += self.batch_size
                    xp = self.X_np[i:i + self.batch_size]
                    yp = self.y_np[i:i + self.batch_size]
                    yield (
                        numpy_to_ort_value(xp, self.device),
                        numpy_to_ort_value(yp, self.device))
        else:
            if b <= 0 or self.batch_size <= 0:
                yield (self.X_ort, self.y_ort, self.w_ort)
            else:
                i = -1
                while N < len(self):
                    i = self._next_iter(i)
                    N += self.batch_size
                    xp = self.X_np[i:i + self.batch_size]
                    yp = self.y_np[i:i + self.batch_size]
                    wp = self.w_np[i:i + self.batch_size]
                    yield (
                        numpy_to_ort_value(xp, self.device),
                        numpy_to_ort_value(yp, self.device),
                        numpy_to_ort_value(wp, self.device))

    def iter_bind(self, bind, names):
        """
        Iterates over the datasets by drawing
        *batch_size* consecutive observations.
        Modifies a bind structure.
        """
        if len(names) not in (3, 4):
            raise NotImplementedError(
                "The dataloader expects three (feature name, label name, "
                "learning rate) or (feature name, label name, sample_weight, "
                "learning rate), not %r." % names)

        n_col_x = self.desc[0][0][1]
        n_col_y = self.desc[1][0][1]
        size_x = self.desc[0][1].itemsize
        size_y = self.desc[1][1].itemsize
        size_w = None if len(self.desc) <= 2 else self.desc[2][1].itemsize

        def local_bind(bind, offset, n):
            # This function assumes the data is contiguous.
            shape_X = (n, n_col_x)
            shape_y = (n, n_col_y)

            try:
                bind.bind_input(
                    names[0], self.device, self.desc[0][1], shape_X,
                    self.X_ort.data_ptr() + offset * n_col_x * size_x)
            except RuntimeError as e:  # pragma: no cover
                raise RuntimeError(
                    "Unable to bind data input (X) %r, device=%r desc=%r "
                    "data_ptr=%r offset=%r n_col_x=%r size_x=%r "
                    "type(bind)=%r" % (
                        names[0], self.device, self.desc[0][1],
                        self.X_ort.data_ptr(), offset, n_col_x, size_x,
                        type(bind))) from e
            try:
                bind.bind_input(
                    names[1], self.device, self.desc[1][1], shape_y,
                    self.y_ort.data_ptr() + offset * n_col_y * size_y)
            except RuntimeError as e:  # pragma: no cover
                raise RuntimeError(
                    "Unable to bind data input (y) %r, device=%r desc=%r "
                    "data_ptr=%r offset=%r n_col_y=%r size_y=%r "
                    "type(bind)=%r" % (
                        names[1], self.device, self.desc[1][1],
                        self.y_ort.data_ptr(), offset, n_col_y, size_y,
                        type(bind))) from e

        def local_bindw(bind, offset, n):
            # This function assumes the data is contiguous.
            shape_w = (n, )

            bind.bind_input(
                names[2], self.device, self.desc[2][1], shape_w,
                self.w_ort.data_ptr() + offset * size_w)

        N = 0
        b = len(self) - self.batch_size
        if self.w_ort is None:
            if b <= 0 or self.batch_size <= 0:
                shape_x = self.desc[0][0]
                local_bind(bind, 0, shape_x[0])
                yield shape_x[0]
            else:
                n = self.batch_size
                i = -1
                while N < len(self):
                    i = self._next_iter(i)
                    N += self.batch_size
                    local_bind(bind, i, n)
                    yield n
        else:
            if b <= 0 or self.batch_size <= 0:
                shape_x = self.desc[0][0]
                local_bind(bind, 0, shape_x[0])
                local_bindw(bind, 0, shape_x[0])
                yield shape_x[0]
            else:
                n = self.batch_size
                i = -1
                while N < len(self):
                    i = self._next_iter(i)
                    N += self.batch_size
                    local_bind(bind, i, n)
                    local_bindw(bind, i, n)
                    yield n

    @property
    def data_np(self):
        "Returns a tuple of the datasets in numpy."
        if self.w_np is None:
            return self.X_np, self.y_np
        return self.X_np, self.y_np, self.w_np

    @property
    def data_ort(self):
        "Returns a tuple of the datasets in onnxruntime C_OrtValue."
        if self.w_ort is None:
            return self.X_ort, self.y_ort
        return self.X_ort, self.y_ort, self.w_ort



# pylint: disable=W0105
"""
@file
@brief Helper for :epkg:`onnxruntime-training`.
"""
import numpy
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from onnxruntime import SessionOptions, InferenceSession, RunOptions
from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611
    OrtValue as C_OrtValue)
from ..utils.onnx_function import function_onnx_graph
from ..utils.onnxruntime_helper import device_to_providers
from ._base_onnx_function import BaseLearningOnnx


class BaseLearningRate(BaseLearningOnnx):
    """
    Class handling the learning rate update after every
    iteration of a gradient. Two methods need to be overwritten
    `init_learning_rate` and `update_learning_rate`. The first one
    starts the loop, the second returns the next one.
    """

    def __init__(self):
        BaseLearningOnnx.__init__(self)
        self.ro_ = RunOptions()

    def _call_iobinding(self, sess, bind):
        sess.run_with_iobinding(bind, self.ro_)

    def init_learning_rate(self):
        """
        Initializes the learning rate at the beginning of the training.
        :return: self
        """
        raise NotImplementedError(
            "This method must be overwritten.")

    def update_learning_rate(self, t):
        """
        Updates the learning rate at the end of an iteration.
        :param t: iteration number
        :return: self
        """
        raise NotImplementedError(
            "This method must be overwritten.")

    @property
    def value(self):
        "Returns the current learning rate."
        raise NotImplementedError(
            "This method must be overwritten.")

    def __repr_extended__(self):
        return ', value=%r' % self.value

    @property
    def needs_grad(self):
        """
        Returns the True if the gradient update needs to retain
        past gradients.
        """
        raise NotImplementedError(
            "This method must be overwritten.")

    def update_weights(self, device, statei, gradienti, batch_size,
                       velocity=None):
        """
        Updates weights based on the algorithm this class
        is setting up.

        :param device: device
        :param statei: current weight
        :param gradienti: gradient
        :param batch_size: batch_size
        :param velocity: same shape as the gradient
        """
        raise NotImplementedError(
            "This method must be overwritten.")

    def loop(self, n=1000):
        """
        Loops over learning rate values, *n* to be precise.
        :param n: number of requested iterations
        :return: iterator
        """
        self.init_learning_rate()
        for i in range(n):
            yield self.value
            self.update_learning_rate(i + 1)

    @staticmethod
    def select(class_name, **kwargs):
        """
        Returns an instance of a given initialized with
        *kwargs*.
        :param class_name: an instance of @see cl BaseLearningRate
            or a string among the following class names (see below),
            it can also be a float and in that case, class
            @see cl LearningRateSGD is used
        :return: instance of @see cl BaseLearningRate

        Possible values for *class_name*:
        * `'SGD'` or `'LearningRateSGD'`: see @see cl LearningRateSGD
        """
        if isinstance(class_name, BaseLearningRate):
            return class_name
        if isinstance(class_name, float):
            return LearningRateSGD(class_name)
        cls = {LearningRateSGD: ['SGD'],
               LearningRateSGDNesterov: ['SGDNesterov', 'Nesterov']}
        for cl, aliases in cls.items():
            if class_name == cl.__class__.__name__ or class_name in aliases:
                return cl(**kwargs)
        raise ValueError(  # pragma: no cover
            "Unexpected class name %r. It should be one of %r." % (
                class_name, list(map(lambda c: c.__name__, cls))))


class LearningRateSGD(BaseLearningRate):
    """
    Implements the learning the same way as
    :class:`sklearn.linear_model.SGDRegressor`.

    :param eta0: initial learning rate for the `'constant'`, `'invscaling'`
        or `'adaptive'` schedules.
    :param alpha: constant that multiplies the regularization term,
        the higher the value, the stronger the regularization.
        Also used to compute the learning rate when set to *learning_rate*
        is set to `'optimal'`.
    :param power_t: exponent for inverse scaling learning rate
    :param learning_rate: learning rate schedule:
        * `'constant'`: `eta = eta0`
        * `'optimal'`: `eta = 1.0 / (alpha * (t + t0))` where *t0* is chosen
            by a heuristic proposed by Leon Bottou, this number is multiplied
            by a constant C to make the first number equal to *eta0*
        * `'invscaling'`: `eta = eta0 / pow(t, power_t)`

    Created attributes:
    * `eta0_`: initial eta0
    * `optimal_init_`: use when `learning_rate=='optimal'`
    * `value_`: value to be returned by property `value`
    """

    def __init__(self, eta0=0.01, alpha=0.0001, power_t=0.25,
                 learning_rate='invscaling'):
        BaseLearningRate.__init__(self)
        if learning_rate not in ('invscaling', 'optimal', 'constant'):
            raise ValueError(
                "Unxepected value for learning_rate=%r." % learning_rate)
        self.eta0 = eta0
        self.alpha = alpha
        self.power_t = power_t
        self.learning_rate = learning_rate.lower()
        self.value_ = None

    @property
    def value(self):
        "Returns the current learning rate."
        if self.value_ is None:
            raise RuntimeError(  # pragma: no cover
                "Method init_learning_rate was never called.")
        return self.value_

    @property
    def needs_grad(self):
        """
        Returns the True if the gradient update needs to retain
        past gradients.
        """
        return False

    def init_learning_rate(self):
        """
        Updates the learning rate at the end of an iteration.
        :return: self
        """
        self.eta0_ = self.eta0
        if self.learning_rate == "optimal":
            typw = numpy.sqrt(1.0 / numpy.sqrt(self.alpha))
            eta0 = typw / max(1.0, (1 + typw) * 2)
            self.optimal_init_ = 1.0 / (eta0 * self.alpha)
            eta = 1. / (self.alpha * self.optimal_init_)
            self.optimal_fact_ = self.eta0 / eta
            self.eta0_ = self.eta0
        else:
            self.eta0_ = self.eta0
        self.value_ = self.eta0_
        return self

    def update_learning_rate(self, t):
        """
        Updates the learning rate at the end of an iteration.
        :param t: iteration number
        :return: self
        """
        eta = self.value_
        if self.learning_rate == "optimal":
            eta = self.optimal_fact_ / (self.alpha * (self.optimal_init_ + t))
        elif self.learning_rate == "invscaling":
            eta = self.eta0_ / numpy.power(t + 1, self.power_t)
        self.value_ = eta
        return self

    def build_onnx_function(self, opset, device, n_tensors):
        so = SessionOptions()
        so.log_severity_level = 4

        self.axpy_onnx_ = function_onnx_graph("axpy")
        self.axpy_sess_ = InferenceSession(
            self.axpy_onnx_.SerializeToString(), so,
            providers=device_to_providers(device))
        self.axpy_sess_binds_ = [
            self.axpy_sess_.io_binding()._iobinding
            for i in range(n_tensors)]
        self.alpha_ = numpy.array(
            [0], dtype=TENSOR_TYPE_TO_NP_TYPE[
                self.axpy_onnx_.graph.input[0].type.tensor_type.elem_type])

    def update_weights(self, n_bind, device, statei, gradienti, batch_size,
                       velocity=None):
        if velocity is not None:
            raise RuntimeError(  # pragma: no cover
                "Velocity must be None for this way of updating weights.")
        if (not hasattr(self, "axpy_onnx_") or
                not hasattr(self, "axpy_sess_binds_")):
            raise RuntimeError(  # pragma: no cover
                "Attributes 'axpy_sess_binds_' or "
                "'axpy_onnx_' is missing. Method "
                "'build_onnx_function' has not been called.")
        bind = self.axpy_sess_binds_[n_bind]
        self._bind_input_ortvalue("X1", bind, gradienti, device, cache=True)
        self._bind_input_ortvalue("X2", bind, statei, device, cache=True)
        self.alpha_[0] = - self.value / batch_size  # pylint: disable=E1130
        ort_alpha = C_OrtValue.ortvalue_from_numpy(self.alpha_, device)
        self._bind_input_ortvalue("alpha", bind, ort_alpha, device, cache=True)
        self._bind_output_ortvalue('Y', bind, statei, cache=True)
        self._call_iobinding(self.axpy_sess_._sess, bind)
        loss = bind.get_outputs()[0]
        return loss


class LearningRateSGDNesterov(LearningRateSGD):
    """
    Implements the learning the same way as
    :class:`sklearn.linear_model.SGDRegressor`.

    :param eta0: initial learning rate for the `'constant'`, `'invscaling'`
        or `'adaptive'` schedules.
    :param alpha: constant that multiplies the regularization term,
        the higher the value, the stronger the regularization.
        Also used to compute the learning rate when set to *learning_rate*
        is set to `'optimal'`.
    :param power_t: exponent for inverse scaling learning rate
    :param learning_rate: learning rate schedule:
        * `'constant'`: `eta = eta0`
        * `'optimal'`: `eta = 1.0 / (alpha * (t + t0))` where *t0* is chosen
            by a heuristic proposed by Leon Bottou, this number is multiplied
            by a constant C to make the first number equal to *eta0*
        * `'invscaling'`: `eta = eta0 / pow(t, power_t)`
    :param momentum: float, default=0.9
        Value of momentum used, must be larger than or equal to 0.
    :param nesterov: bool, default=True
        Whether to use nesterov's momentum or not. Use nesterov's if True
        Not using nesterov is equivalent to class @see cl LearningRateSGD.

    Created attributes:
    * `eta0_`: initial eta0
    * `optimal_init_`: use when `learning_rate=='optimal'`
    * `value_`: value to be returned by property `value`

    ::

        updates = [
            self.momentum * velocity - self.learning_rate * grad
            for velocity, grad in zip(self.velocities, grads)]
        self.velocities = updates

        if self.nesterov:
            updates_nesterov = [
                self.momentum * velocity - self.learning_rate * grad
                for velocity, grad in zip(self.velocities, grads)]
            return updates, updates_nesterov    --> new gradient and velocities
        else:
            return updates                      --> new gradient
    """

    def __init__(self, eta0=0.01, alpha=0.0001, power_t=0.25,
                 learning_rate='invscaling', momentum=0.9, nesterov=True):
        LearningRateSGD.__init__(
            self, eta0=eta0, alpha=alpha, power_t=power_t,
            learning_rate=learning_rate)
        self.momentum = momentum
        self.nesterov = nesterov

    @property
    def needs_grad(self):
        """
        Returns the True if the gradient update needs to retain
        past gradients.
        """
        return True

    def init_learning_rate(self):
        """
        Updates the learning rate at the end of an iteration.
        :return: self
        """
        return LearningRateSGD.init_learning_rate(self)

    def update_learning_rate(self, t):
        """
        Updates the learning rate at the end of an iteration.
        :param t: iteration number
        :return: self
        """
        return LearningRateSGD.update_learning_rate(self, t)

    def build_onnx_function(self, opset, device, n_tensors):
        so = SessionOptions()
        so.log_severity_level = 4

        # axpyw
        if self.nesterov:
            self.axpyw_onnx_ = function_onnx_graph("axpyw2")
        else:
            self.axpyw_onnx_ = function_onnx_graph("axpyw")
        self.axpyw_sess_ = InferenceSession(
            self.axpyw_onnx_.SerializeToString(), so,
            providers=device_to_providers(device))
        self.axpyw_sess_binds_ = [
            self.axpyw_sess_.io_binding()._iobinding
            for n in range(n_tensors)]

        self.alpha_ = numpy.array(
            [0], dtype=TENSOR_TYPE_TO_NP_TYPE[
                self.axpyw_onnx_.graph.input[0].type.tensor_type.elem_type])
        self.beta_ = numpy.array(
            [0], dtype=TENSOR_TYPE_TO_NP_TYPE[
                self.axpyw_onnx_.graph.input[0].type.tensor_type.elem_type])

    def update_weights(self, n_bind, device, statei, gradienti, batch_size,
                       velocity=None):
        if (not hasattr(self, "axpyw_onnx_") or
                not hasattr(self, "axpyw_sess_binds_")):
            raise RuntimeError(  # pragma: no cover
                "Attributes 'axpyw_sess_binds_' or "
                "'axpyw_onnx_' is missing. Method "
                "'build_onnx_function' has not been called.")
        if velocity is None:
            raise RuntimeError(  # pragma: no cover
                "Velocity must not be None for this way of updating weights.")
        bind = self.axpyw_sess_binds_[n_bind]
        self._bind_input_ortvalue("X1", bind, gradienti, device, cache=True)
        self._bind_input_ortvalue("X2", bind, statei, device, cache=True)
        self._bind_input_ortvalue("G", bind, velocity, device, cache=True)
        self.alpha_[0] = - self.value / batch_size  # pylint: disable=E1130
        self.beta_[0] = self.momentum
        ort_alpha = C_OrtValue.ortvalue_from_numpy(self.alpha_, device)
        ort_beta = C_OrtValue.ortvalue_from_numpy(self.beta_, device)
        self._bind_input_ortvalue("alpha", bind, ort_alpha, device, cache=True)
        self._bind_input_ortvalue("beta", bind, ort_beta, device, cache=True)
        self._bind_output_ortvalue('Y', bind, statei, cache=True)
        self._bind_output_ortvalue('Z', bind, velocity, cache=True)
        self._call_iobinding(self.axpyw_sess_._sess, bind)
        return bind.get_outputs()  # loss, velocity



# pylint: disable=C0415
"""
@file
@brief Implements command line ``python -m onnxcustom <command> <args>``.
"""


def main(args, fLOG=print):
    """
    Implements ``python -m onnxcustom <command> <args>``.

    :param args: command line arguments
    :param fLOG: logging function
    """
    from pyquickhelper.cli import cli_main_helper
    try:
        from . import check
        from .cli.profiling import nvprof2json
    except ImportError:  # pragma: no cover
        from onnxcustom import check
        from onnxcustom.cli.profiling import nvprof2json

    fcts = dict(nvprof2json=nvprof2json, check=check)
    return cli_main_helper(fcts, args=args, fLOG=fLOG)


if __name__ == "__main__":
    import sys  # pragma: no cover
    main(sys.argv[1:])  # pragma: no cover



"""
@file
@brief Base class for @see cl BaseEstimator and @see cl BaseOnnxFunction.
"""
import os
import inspect
import warnings


class BaseOnnxClass:
    """
    Bases class with common functions to handle attributes
    in classes owning ONNX graphs.
    """

    @classmethod
    def _get_param_names(cls):
        "Extracts all parameters to serialize."
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        init_signature = inspect.signature(init)
        parameters = [
            p for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD]
        return [(p.name, p.default) for p in parameters]

    def save_onnx_graph(self, folder, prefix=None, suffix=None):
        """
        Saves all ONNX files stored in this class.

        :param folder: folder where to save (it must exists) or
            ``bytes`` if the onnx graph must be returned as bytes,
            not files
        :param prefix: suffix to add to the name
        :param suffix: suffix to add to the name
        :return: list of saved files (dictionary
            `{ attribute: filename or dictionary }`)

        The function raises a warning if a file already exists.
        The function uses class name, attribute names to compose
        file names. It shortens them for frequent classes.

        * 'Learning' -> 'L'
        * 'OrtGradient' -> 'Grad'
        * 'ForwardBackward' -> 'FB'

        .. runpython::
            :showcode:

            import io
            import numpy
            import onnx
            from sklearn.datasets import make_regression
            from sklearn.model_selection import train_test_split
            from sklearn.linear_model import LinearRegression
            from skl2onnx import to_onnx
            from mlprodict.plotting.text_plot import onnx_simple_text_plot
            from onnxcustom.training.optimizers_partial import (
                OrtGradientForwardBackwardOptimizer)
            from onnxcustom.training.sgd_learning_rate import (
                LearningRateSGDNesterov)
            from onnxcustom.training.sgd_learning_penalty import (
                ElasticLearningPenalty)


            def walk_through(obj, prefix="", only_name=True):
                for k, v in obj.items():
                    if isinstance(v, dict):
                        p = prefix + "." + k if prefix else k
                        walk_through(v, prefix=p, only_name=only_name)
                    elif only_name:
                        name = "%s.%s" % (prefix, k) if prefix else k
                        print('+', name)
                    else:
                        name = "%s.%s" % (prefix, k) if prefix else k
                        print('\\n++++++', name)
                        print()
                        bf = io.BytesIO(v)
                        onx = onnx.load(bf)
                        print(onnx_simple_text_plot(onx))


            X, y = make_regression(  # pylint: disable=W0632
                100, n_features=3, bias=2, random_state=0)
            X = X.astype(numpy.float32)
            y = y.astype(numpy.float32)
            X_train, _, y_train, __ = train_test_split(X, y)
            reg = LinearRegression()
            reg.fit(X_train, y_train)
            reg.coef_ = reg.coef_.reshape((1, -1))
            opset = 15
            onx = to_onnx(reg, X_train, target_opset=opset,
                          black_op={'LinearRegressor'})
            inits = ['coef', 'intercept']

            train_session = OrtGradientForwardBackwardOptimizer(
                onx, inits,
                learning_rate=LearningRateSGDNesterov(
                    1e-4, nesterov=False, momentum=0.9),
                learning_penalty=ElasticLearningPenalty(l1=1e-3, l2=1e-4),
                warm_start=False, max_iter=100, batch_size=10)

            onxs = train_session.save_onnx_graph(bytes)

            print("+ all onnx graphs")
            walk_through(onxs, only_name=True)
            walk_through(onxs, only_name=False)
        """
        repls = {'Learning': 'L', 'OrtGradient': 'Grad',
                 'ForwardBackward': 'FB'}
        if prefix is None:
            prefix = ''
        if suffix is None:
            suffix = ''
        if isinstance(folder, str) and not os.path.exists(folder):
            raise FileNotFoundError(  # pragma: no cover
                "Folder %r does not exist." % folder)
        saved = {}
        for k, v in self.__dict__.items():
            if hasattr(v, "SerializeToString"):
                if isinstance(folder, str):
                    name = "%s%s%s.%s.onnx" % (
                        prefix, self.__class__.__name__, suffix, k)
                    for a, b in repls.items():
                        name = name.replace(a, b)
                    filename = os.path.join(folder, name)
                    if os.path.exists(filename):
                        warnings.warn(  # pragma: no cover
                            "Filename %r already exists." % filename)
                    with open(filename, "wb") as f:
                        f.write(v.SerializeToString())
                    saved[k] = filename
                else:
                    saved[k] = v.SerializeToString()
            elif hasattr(v, "save_onnx_graph"):
                saved[k] = v.save_onnx_graph(
                    folder, prefix=prefix, suffix="%s.%s" % (suffix, k))
        return saved



"""
@file
@brief Optimizer with :epkg:`onnxruntime-training`.
"""
import numpy
from onnxruntime import (  # pylint: disable=E0611
    TrainingParameters, SessionOptions, TrainingSession)
from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611
    OrtValue as C_OrtValue, SessionIOBinding as C_IOBinding)
from ..utils.onnxruntime_helper import (
    numpy_to_ort_value, device_to_providers)
from .data_loader import OrtDataLoader
from .excs import ConvergenceError, EvaluationError
from ._base_estimator import BaseEstimator


class OrtGradientOptimizer(BaseEstimator):
    """
    Implements a simple :epkg:`Stochastic Gradient Descent`
    with :epkg:`onnxruntime-training`.

    :param model_onnx: ONNX graph used to train
    :param weights_to_train: names of initializers to be optimized
    :param loss_output_name: name of the loss output
    :param max_iter: number of training iterations
    :param training_optimizer_name: optimizing algorithm
    :param batch_size: batch size (see class *DataLoader*)
    :param learning_rate: a name or a learning rate instance or a float,
        see module :mod:`onnxcustom.training.sgd_learning_rate`
    :param device: device as :epkg:`C_OrtDevice` or a string
        representing this device
    :param warm_start: when set to True, reuse the solution of the previous
        call to fit as initialization, otherwise, just erase the previous
        solution.
    :param verbose: use :epkg:`tqdm` to display the training progress
    :param validation_every: validation with a test set every
        *validation_every* iterations
    :param saved_gradient: if not None, a filename,
        the optimizer saves the gradient into it
    :param sample_weight_name: name of the sample weight input

    Once initialized, the class creates the attribute
    `train_session_` which holds an instance of :ref:`l-ort-training-session`.

    See example :ref:`l-orttraining-nn-gpu`.
    """

    def __init__(self, model_onnx, weights_to_train, loss_output_name='loss',
                 max_iter=100, training_optimizer_name='SGDOptimizer',
                 batch_size=10, learning_rate='SGD',
                 device='cpu', warm_start=False, verbose=0,
                 validation_every=0.1, saved_gradient=None,
                 sample_weight_name="weight"):
        BaseEstimator.__init__(self, learning_rate, device)
        self.model_onnx = model_onnx
        self.batch_size = batch_size
        self.weights_to_train = weights_to_train
        self.loss_output_name = loss_output_name
        self.training_optimizer_name = training_optimizer_name
        self.verbose = verbose
        self.max_iter = max_iter
        self.warm_start = warm_start
        self.saved_gradient = saved_gradient
        self.sample_weight_name = sample_weight_name
        if validation_every < 1:
            self.validation_every = int(self.max_iter * validation_every)
        else:
            self.validation_every = validation_every  # pragma: no cover
        if self.learning_rate.needs_grad:
            raise NotImplementedError(
                "Any weight update involving past gradient is "
                "not implemented (class %r)."
                "" % self.learning_rate.__class__.__name__)

    def fit(self, X, y, sample_weight=None, X_val=None, y_val=None,
            use_numpy=False):
        """
        Trains the model.

        :param X: features
        :param y: expected output
        :param sample_weight: sample weight if any
        :param X_val: evaluation dataset
        :param y_val: evaluation dataset
        :param use_numpy: if True, slow iterator using numpy,
            otherwise, minimizes copy
        :return: self
        """
        input_names = [i.name for i in self.model_onnx.graph.input]
        if ((len(input_names) == 2 and sample_weight is not None) or
                (len(input_names) == 3 and sample_weight is None)):
            raise RuntimeError(  # pragma: no cover
                "Number of inputs should be 2 if sample_weight is None "
                "or 3 if not None but it is %d." % len(input_names))
        self.train_session_ = self._create_training_session(
            self.model_onnx, self.weights_to_train,
            loss_output_name=self.loss_output_name,
            training_optimizer_name=self.training_optimizer_name,
            device=self.device)

        if not self.warm_start:
            state = self.get_state()
            new_state = {}
            for k, v in state.items():
                if len(v.shape) > 0:
                    new_state[k] = numpy.random.randn(*v.shape).astype(v.dtype)
                else:
                    f = numpy.random.randn(1)
                    f = f.astype(v.dtype)
                    new_state[k] = f
            self.set_state(new_state)

        data_loader = OrtDataLoader(
            X, y, sample_weight=sample_weight,
            batch_size=self.batch_size, device=self.device)
        if X_val is not None:
            data_loader_val = OrtDataLoader(
                X_val, y_val, batch_size=X_val.shape[0], device=self.device,
                random_iter=False)
        else:
            data_loader_val = None

        self.learning_rate.init_learning_rate()
        self.input_names_ = [i.name for i in self.train_session_.get_inputs()]
        self.output_names_ = [
            o.name for o in self.train_session_.get_outputs()]
        self.loss_index_ = self.output_names_.index(self.loss_output_name)

        bind = self.train_session_.io_binding()._iobinding

        if self.verbose > 0:  # pragma: no cover
            from tqdm import tqdm  # pylint: disable=C0415
            loop = tqdm(range(self.max_iter))
        else:
            loop = range(self.max_iter)

        train_losses = []
        val_losses = []
        lr = self.learning_rate.value
        for it in loop:
            lr_alive = numpy.array([lr / self.batch_size], dtype=numpy.float32)
            ort_lr = numpy_to_ort_value(lr_alive, self.device)
            loss = self._iteration(data_loader, ort_lr,
                                   bind, use_numpy=use_numpy,
                                   sample_weight=sample_weight is not None)
            lr = self.learning_rate.update_learning_rate(it).value
            if self.verbose > 1:  # pragma: no cover
                loop.set_description(
                    "loss=%1.3g lr=%1.3g "  # pylint: disable=E1307
                    "lrn=%1.3g" % (
                        loss, lr, lr_alive[0]))
            train_losses.append(loss)
            if (data_loader_val is not None and
                    (it + 1) % self.validation_every == 0):
                val_losses.append(self._evaluation(data_loader_val, bind))
        self.train_losses_ = train_losses
        self.validation_losses_ = (
            None if data_loader_val is None else val_losses)
        self.trained_coef_ = self.train_session_.get_state()
        return self

    def _bind_input_ortvalue(self, name, bind, c_ortvalue):
        """
        Binds :epkg:`C_OrtValue` to the structure used by
        :epkg:`InferenceSession` to run inference.

        :param name: str
        :param bind: python structure
        :param c_ortvalue: C structure for OrtValue (:epkg:`C_OrtValue`),
            it can be also a numpy array
        """
        if not isinstance(bind, C_IOBinding):
            raise TypeError(  # pragma: no cover
                "Unexpected type %r." % type(bind))
        if isinstance(c_ortvalue, C_OrtValue):
            bind.bind_ortvalue_input(name, c_ortvalue)
        elif isinstance(c_ortvalue, numpy.ndarray):
            # This fails on linux with int64.
            bind.bind_input(
                name, self.device, c_ortvalue.dtype, c_ortvalue.shape,
                c_ortvalue.__array_interface__['data'][0])
        else:
            raise TypeError(  # pragma: no cover
                "Unable to bind type %r for name %r." % (
                    type(c_ortvalue), name))

    def _iteration(self, data_loader, ort_lr, bind, use_numpy, sample_weight):
        actual_losses = []

        bind.bind_output('loss', self.device)

        if use_numpy:
            # onnxruntime does not copy the data, so the numpy
            # array must remain alive all along the iteration
            lr_alive = ort_lr.numpy()
            idx = 3 if sample_weight else 2
            self._bind_input_ortvalue(
                self.input_names_[idx], bind, lr_alive)

            # Slow iterations.
            for it in data_loader.iter_numpy():
                if len(it) == 2:
                    data, target = it
                    self._bind_input_ortvalue(
                        self.input_names_[0], bind, data)
                    self._bind_input_ortvalue(
                        self.input_names_[1], bind, target)
                else:
                    data, target, weight = it
                    self._bind_input_ortvalue(
                        self.input_names_[0], bind, data)
                    self._bind_input_ortvalue(
                        self.input_names_[1], bind, target)
                    self._bind_input_ortvalue(
                        self.input_names_[2], bind, weight)

                self.train_session_._sess.run_with_iobinding(bind, None)
                loss = bind.get_outputs()[0].numpy()
                if numpy.isinf(loss) or numpy.isnan(loss):
                    raise ConvergenceError(
                        "Loss is nan, learning_rate=%r, "
                        "the gradient descent has failed "
                        "(past losses=%r)." % (
                            ort_lr.numpy(),
                            [float(v[0]) for v in (
                                actual_losses if len(actual_losses) < 5
                                else actual_losses[-5:])]))
                actual_losses.append(loss / data.shape[0])
        else:
            idx = 3 if sample_weight else 2
            self._bind_input_ortvalue(self.input_names_[idx], bind, ort_lr)

            # Fast iterations
            # Slow iterations.
            for batch_size in data_loader.iter_bind(bind, self.input_names_):
                self.train_session_._sess.run_with_iobinding(bind, None)
                # We copy the predicted output as well which is not needed.
                loss = bind.get_outputs()[0].numpy()
                if numpy.isinf(loss) or numpy.isnan(loss):
                    raise ConvergenceError(
                        "Loss is nan or infinite, learning_rate=%r, "
                        "the gradient descent has failed "
                        "(past losses=%r)." % (
                            ort_lr.numpy(),
                            [float(v[0]) for v in (
                                actual_losses if len(actual_losses) < 5
                                else actual_losses[-5:])]))
                actual_losses.append(loss / batch_size)

        return numpy.array(actual_losses).mean()

    def _evaluation(self, data_loader, bind):
        lr_alive = numpy.array([0], dtype=numpy.float32)
        self._bind_input_ortvalue(self.input_names_[2], bind, lr_alive)
        bind.bind_output('loss', self.device)

        actual_losses = []
        for batch_size in data_loader.iter_bind(bind, self.input_names_):
            self.train_session_._sess.run_with_iobinding(bind, None)
            outputs = bind.copy_outputs_to_cpu()
            if numpy.isinf(outputs[0]) or numpy.isnan(outputs[0]):
                raise EvaluationError(  # pragma: no cover
                    "Loss is nan or infinite (%r), "
                    "evaluation has failed." % outputs[0])
            actual_losses.append(outputs[0] / batch_size)
        return numpy.array(actual_losses).sum() / len(data_loader)

    def _create_training_session(
            self, training_onnx, weights_to_train,
            loss_output_name='loss',
            training_optimizer_name='SGDOptimizer',
            device='cpu'):
        """
        Creates an instance of :epkg:`TrainingSession`.

        :param training_onnx: an ONNX graph with a loss function
        :param weights_to_train: list of initializer names to optimize
        :param loss_output_name: output name for the loss
        :param training_optimizer_name: optimizer name
        :param device: one :epkg:`C_OrtDevice` or a string
        :return: an instance of :epkg:`TrainingSession`
        """
        if training_optimizer_name != 'SGDOptimizer':
            raise NotImplementedError(
                "Only the SGDOptimizer is implemented not %r."
                "" % training_optimizer_name)
        ort_parameters = TrainingParameters()
        ort_parameters.loss_output_name = loss_output_name
        ort_parameters.use_mixed_precision = False
        # ort_parameters.world_rank = -1
        # ort_parameters.world_size = 1
        # ort_parameters.gradient_accumulation_steps = 1
        # ort_parameters.allreduce_post_accumulation = False
        # ort_parameters.deepspeed_zero_stage = 0
        # ort_parameters.enable_grad_norm_clip = False
        # ort_parameters.set_gradients_as_graph_outputs = False
        # ort_parameters.use_memory_efficient_gradient = False
        # ort_parameters.enable_adasum = False
        if self.saved_gradient is not None:
            name = self.saved_gradient
            name2 = name + ".training.onnx"
            ort_parameters.model_with_gradient_graph_path = name
            ort_parameters.model_with_training_graph_path = name2

        output_types = {}
        for output in training_onnx.graph.output:
            output_types[output.name] = output.type.tensor_type

        ort_parameters.weights_to_train = set(weights_to_train)
        ort_parameters.training_optimizer_name = training_optimizer_name
        # ort_parameters.lr_params_feed_name = lr_params_feed_name

        ort_parameters.optimizer_attributes_map = {
            name: {} for name in weights_to_train}
        ort_parameters.optimizer_int_attributes_map = {
            name: {} for name in weights_to_train}

        session_options = SessionOptions()
        session_options.log_severity_level = 4
        session_options.log_verbosity_level = 4
        # session_options.use_deterministic_compute = True

        providers = device_to_providers(self.device)
        session = TrainingSession(
            training_onnx.SerializeToString(), ort_parameters, session_options,
            providers=providers)

        return session

    def get_state(self):
        """
        Returns the trained weights.
        """
        if not hasattr(self, 'train_session_'):
            if hasattr(self, 'trained_coef_'):
                return self.trained_coef_
            raise AttributeError("Method fit must be called before.")
        return self.train_session_.get_state()

    def set_state(self, state):
        """
        Changes the trained weights.
        """
        if not hasattr(self, 'train_session_'):
            raise AttributeError(  # pragma: no cover
                "Method fit must be called before.")
        return self.train_session_.load_state(state)



"""
@file
@brief Optimizer with :epkg:`onnxruntime-training`.
"""
import inspect
from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611
    OrtDevice as C_OrtDevice)
from ..utils.onnxruntime_helper import (
    get_ort_device, ort_device_to_string)
from ._base import BaseOnnxClass
from ._base_onnx_function import BaseLearningOnnx
from .sgd_learning_rate import BaseLearningRate


class BaseEstimator(BaseOnnxClass):
    """
    Base class for optimizers.
    Implements common methods such `__repr__`.

    :param learning_rate: learning rate class,
        see module :mod:`onnxcustom.training.sgd_learning_rate`
    :param device: device as :epkg:`C_OrtDevice` or a string
        representing this device
    """

    def __init__(self, learning_rate, device):
        self.learning_rate = BaseLearningRate.select(learning_rate)
        self.device = get_ort_device(device)

    @classmethod
    def _get_param_names(cls):
        "Extracts all parameters to serialize."
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        init_signature = inspect.signature(init)
        parameters = [
            p for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD]
        return [(p.name, p.default) for p in parameters]

    def __repr__(self):
        "Usual."
        param = self._get_param_names()
        ps = []
        for k, v in param:
            if k not in self.__dict__:
                continue  # pragma: no cover
            ov = getattr(self, k)
            if isinstance(ov, BaseLearningOnnx):
                ps.append("%s=%s" % (k, repr(ov)))
            elif isinstance(ov, C_OrtDevice):
                ps.append("%s=%r" % (k, ort_device_to_string(ov)))
            elif v is not inspect._empty or ov != v:
                ro = repr(ov)
                if len(ro) > 50 or "\n" in ro:
                    ro = ro[:10].replace("\n", " ") + "..."
                    ps.append("%s=%r" % (k, ro))
                else:
                    ps.append("%s=%r" % (k, ov))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(ps))

    def __getstate__(self):
        "Removes any non pickable attribute."
        atts = [k for k in self.__dict__ if not k.endswith('_')]
        if hasattr(self, 'trained_coef_'):
            atts.append('trained_coef_')
        state = {att: getattr(self, att) for att in atts}
        state['device'] = ort_device_to_string(state['device'])
        return state

    def __setstate__(self, state):
        "Restores any non pickable attribute."
        for att, v in state.items():
            setattr(self, att, v)
        self.device = get_ort_device(self.device)
        return self



"""
@file
@brief Exceptions.
"""


class ConvergenceError(RuntimeError):
    """
    Raised when a learning algorithm failed
    to converge.
    """
    pass


class EvaluationError(RuntimeError):
    """
    Raised when an evaluation failed.
    """
    pass


class ProviderError(RuntimeError):
    """
    Raised when an input is not on the expected device (CPU, GPU).
    """
    pass



"""
@file
@brief Optimizer with :epkg:`onnxruntime-training` forward backward training.
"""
import logging
import numpy
from onnxruntime import InferenceSession, SessionOptions
from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611
    OrtValue as C_OrtValue)
from ..utils.onnx_helper import get_onnx_opset, proto_type_to_dtype
from ..utils.onnxruntime_helper import (
    device_to_providers, numpy_to_ort_value)
from ..utils.onnx_function import function_onnx_graph
from ..utils.print_helper import str_ortvalue
from ..utils.orttraining_helper import get_train_initializer
from .ortgradient import OrtGradientForwardBackward
from ._base_estimator import BaseEstimator
from .sgd_learning_loss import BaseLearningLoss
from .sgd_learning_penalty import BaseLearningPenalty
from .data_loader import OrtDataLoader
from .excs import ConvergenceError


class OrtGradientForwardBackwardOptimizer(BaseEstimator):
    """
    Implements a simple :epkg:`Stochastic Gradient Descent`
    with :epkg:`onnxruntime-training`. It leverages class
    @see class OrtGradientForwardBackward.

    :param model_onnx: ONNX graph used to train
    :param weights_to_train: names of initializers to be optimized,
        if None, function @see fn get_train_initialize returns
        the list of float iniitializer
    :param loss_output_name: name of the loss output
    :param max_iter: number of training iterations
    :param training_optimizer_name: optimizing algorithm
    :param batch_size: batch size (see class *DataLoader*)
    :param learning_rate: a name or a learning rate instance or a float,
        see module :mod:`onnxcustom.training.sgd_learning_rate`
    :param device: device as :epkg:`C_OrtDevice` or a string
        representing this device
    :param warm_start: when set to True, reuse the solution of the previous
        call to fit as initialization, otherwise, just erase the previous
        solution.
    :param learning_loss: loss function (see below)
    :param verbose: use :epkg:`tqdm` to display the training progress
    :param validation_every: validation with a test set every
        *validation_every* iterations
    :param enable_logging: enable logging (mostly for debugging puporse
        as it slows down the training)
    :param weight_name: if not None, the class assumes it is trained
        with training weight
    :param learning_penalty: weight penalty, None, or instance of
        @see cl BaseLearningPenalty

    *learning_rate* can be any instance of @see cl BaseLearningRate or
    a nick name in the following list as specified in
    :meth:`BaseLearningRate.select
    <onnxcustom.training.sgd_learning_loss.BaseLearningRate.select>`.

    *learning_loss* can be any instance of @see cl BaseLearningLoss or
    a nick name in the following list as specified in
    :meth:`BaseLearningLoss.select
    <onnxcustom.training.sgd_loss.BaseLearningLoss.select>`.
    """

    def __init__(self, model_onnx, weights_to_train=None,
                 loss_output_name='loss', max_iter=100,
                 training_optimizer_name='SGDOptimizer',
                 batch_size=10, learning_rate='SGD',
                 device='cpu', warm_start=False, verbose=0,
                 validation_every=0.1, learning_loss="square_error",
                 enable_logging=False, weight_name=None,
                 learning_penalty=None):
        if weights_to_train is None:
            weights_to_train = list(get_train_initializer(model_onnx))
        BaseEstimator.__init__(self, learning_rate, device)
        self.model_onnx = model_onnx
        self.batch_size = batch_size
        self.weights_to_train = weights_to_train
        self.loss_output_name = loss_output_name
        self.training_optimizer_name = training_optimizer_name
        self.verbose = verbose
        self.max_iter = max_iter
        self.warm_start = warm_start
        self.learning_loss = BaseLearningLoss.select(learning_loss)
        self.learning_penalty = BaseLearningPenalty.select(learning_penalty)
        self.enable_logging = enable_logging
        self.weight_name = weight_name
        if validation_every < 1:
            self.validation_every = int(self.max_iter * validation_every)
        else:
            self.validation_every = validation_every  # pragma: no cover
        self.build_onnx_function()

    @property
    def needs_grad(self):
        """
        Returns the True if the gradient update needs to retain
        past gradients.
        """
        return self.learning_rate.needs_grad

    def __getstate__(self):
        "Removes any non pickable attribute."
        state = BaseEstimator.__getstate__(self)
        for att in ['train_state_', 'train_grad_state_']:
            if hasattr(self, att):
                train_state = []
                for v in self.get_state():
                    if v is None:
                        train_state.append(v)
                    else:
                        train_state.append(v.numpy())
                state[att[:-1]] = train_state
        return state

    def __setstate__(self, state):
        "Restores any non pickable attribute."
        popped = {}
        for att in ['train_state', 'train_grad_state']:
            if att in state:
                popped[att] = state.pop(att)
        BaseEstimator.__setstate__(self, state)
        for k, v in popped.items():
            if k == 'train_state':
                self.set_state(v, check_trained=False, kind='weight')
            elif k == 'train_grad_state':
                self.set_state(v, check_trained=False, kind='grad')
            else:
                raise ValueError(  # pragma: no cover
                    "Unexpected key state %r." % k)
        self.build_onnx_function()
        return self

    def _get_att_state(self, kind):
        if kind == 'weight':
            return 'train_state_'
        if kind == 'grad':
            return 'train_grad_state_'
        raise ValueError(  # pragma: no cover
            "Unexpected kind=%r." % kind)

    def get_full_state(self, kind='weight'):
        """
        Returns the trained weights.
        """
        if isinstance(kind, list):
            return [self.get_full_state(kind=k) for k in kind]
        att = self._get_att_state(kind)
        if not hasattr(self, att):
            raise AttributeError(  # pragma: no cover
                "Method fit must be called before.")
        return getattr(self, att)

    def get_state(self, kind='weight'):
        """
        Returns the trained weights.
        """
        att = self._get_att_state(kind)
        if not hasattr(self, att):
            raise AttributeError("Method fit must be called before.")
        if getattr(self, att, None) is None:
            raise RuntimeError(  # pragma: no cover
                "No attribute %r available (None)." % att)
        if self.weights_to_train is None:
            raise RuntimeError(  # pragma: no cover
                "Unexpected self.weights_to_train (None).")
        value = getattr(self, att)
        n = len(value) - len(self.weights_to_train)
        return value[n:]

    def set_state(self, state, check_trained=True, kind='weight', zero=False):
        """
        Changes the trained weights.
        """
        if check_trained and not hasattr(self, 'train_session_'):
            raise AttributeError(  # pragma: no cover
                "Method fit must be called before.")
        state_ = []
        state_numpy_ = []
        for i, v in enumerate(state):
            if v is None:
                state_.append(None)
                state_numpy_.append(None)
            elif isinstance(v, numpy.ndarray):
                if zero:
                    v = numpy.zeros(v.shape, dtype=v.dtype)
                ortvalue = numpy_to_ort_value(v, self.device)
                state_.append(ortvalue)
                # The numpy container must be retained as the ortvalue
                # just borrows the pointer.
                state_numpy_.append(v)
            elif isinstance(v, C_OrtValue):
                if zero:
                    v = self.zero_sess_.run_with_ort_values(['Y'], {'X': v})
                state_.append(v)
                state_numpy_.append(None)
            else:
                raise TypeError(  # pragma: no cover
                    "Unexpected type %r for state %r." % (
                        type(v), i))
        att = self._get_att_state(kind)
        setattr(self, att, state_)
        setattr(self, att + "numpy_", state_numpy_)

    def build_onnx_function(self):
        """
        Creates ONNX graph and *InferenceSession* related to
        any operations applying on *OrtValue*.
        """
        opset = get_onnx_opset(self.model_onnx)
        so = SessionOptions()
        so.log_severity_level = 4

        n = len(self.weights_to_train)

        # loss_grad
        self.learning_loss.build_onnx_function(
            opset, self.device, self.weight_name)

        # weight update
        self.learning_rate.build_onnx_function(opset, self.device, n)

        # regularization
        self.learning_penalty.build_onnx_function(opset, self.device, n)

        # zero
        self.zero_onnx_ = function_onnx_graph("zero")
        self.zero_sess_ = InferenceSession(
            self.zero_onnx_.SerializeToString(), so,
            providers=device_to_providers(self.device))

        # logging
        if self.enable_logging:
            self._logger = logging.getLogger("onnxcustom")
        else:
            self._logger = None

    def fit(self, X, y, sample_weight=None,
            X_val=None, y_val=None, use_numpy=False):
        """
        Trains the model.

        :param X: features
        :param y: expected output
        :param sample_weight: training weight or None
        :param X_val: evaluation dataset
        :param y_val: evaluation dataset
        :param use_numpy: if True, slow iterator using numpy,
            otherwise, minimizes copy
        :return: self
        """
        if self.training_optimizer_name != 'SGDOptimizer':
            raise NotImplementedError(
                "Only the SGDOptimizer is implemented not %r."
                "" % self.training_optimizer_name)
        logger = self._logger

        session_function = self._create_training_session(
            self.model_onnx, self.weights_to_train,
            device=self.device)
        self.train_session_ = session_function[0]
        self.train_function_ = session_function[1]

        self.input_names_ = self.train_session_.cls_type_._grad_input_names
        self.output_names_ = self.train_session_.cls_type_._bw_fetches_names
        weights_to_train = self.train_session_.weights_to_train

        if logger is not None:
            logger.info(
                "[OrtGradientForwardBackwardOptimizer.fit] "
                "input_names=%r", self.input_names_)
            logger.info(
                "[OrtGradientForwardBackwardOptimizer.fit] "
                "output_names=%r", self.output_names_)
            logger.info(
                "[OrtGradientForwardBackwardOptimizer.fit] "
                "weights_to_train=%r", self.weights_to_train)
            logger.info(
                "[OrtGradientForwardBackwardOptimizer.fit] "
                "device=%r", self.device)
            if logger is not None:
                logger.info(
                    "[OrtGradientForwardBackwardOptimizer.fit] "
                    "warm_start=%r", self.warm_start)

        if not hasattr(self, 'state_'):
            self.set_state([
                self.train_session_.get_initializer(name, exc=False)
                for name in self.input_names_])
        if self.needs_grad and not hasattr(self, 'state_grad_'):
            self.set_state([
                self.train_session_.get_initializer(name, exc=False)
                for name in self.input_names_],
                kind='grad', zero=True)
        if not self.warm_start:
            state = self.get_full_state()
            if len(state) != len(self.input_names_):
                raise RuntimeError(  # pragma: no cover
                    "Length mismatch %r != %r." % (
                        len(state), len(self.input_names_)))
            new_state = []
            for iv, v in enumerate(state):
                if v is None:
                    new_state.append(v)
                else:
                    if not isinstance(v, C_OrtValue):
                        raise RuntimeError(  # pragma: no cover
                            "Unexpected type %r (state[%d])." % (
                                type(v), iv))
                    dtype = proto_type_to_dtype(
                        v.proto_type()
                        if hasattr(v, 'proto_type')
                        else v.data_type())
                    if len(v.shape()) > 0:
                        new_state.append(
                            numpy.random.randn(*v.shape()).astype(dtype))
                    else:
                        new_state.append(
                            numpy.random.randn(1).astype(dtype))
            self.set_state(new_state)
            if self.needs_grad:
                self.set_state(new_state, kind='grad', zero=True)

        data_loader = OrtDataLoader(
            X, y, sample_weight, batch_size=self.batch_size,
            device=self.device)
        if X_val is not None:
            data_loader_val = OrtDataLoader(
                X_val, y_val, batch_size=X_val.shape[0], device=self.device,
                random_iter=False)
        else:
            data_loader_val = None

        self.learning_rate.init_learning_rate()

        if self.verbose > 0:  # pragma: no cover
            from tqdm import tqdm  # pylint: disable=C0415
            loop = tqdm(range(self.max_iter))
        else:
            loop = range(self.max_iter)

        train_losses = []
        val_losses = []
        kinds = ['weight', 'grad'] if self.needs_grad else ['weight']
        for it in loop:
            loss = self._iteration(
                data_loader, self.get_full_state(kind=kinds),
                len(weights_to_train))
            lr = self.learning_rate.update_learning_rate(it).value
            if self.verbose > 1:  # pragma: no cover
                loop.set_description(
                    "loss=%1.3g lr=%1.3g" % (  # pylint: disable=E1101,E1307
                        loss, lr))  # pylint: disable=E1101,E1307
            train_losses.append(loss)
            if (data_loader_val is not None and
                    (it + 1) % self.validation_every == 0):
                val_losses.append(
                    self._evaluation(data_loader_val, self.get_full_state()))
        self.train_losses_ = train_losses
        self.validation_losses_ = (
            None if data_loader_val is None else val_losses)

        if logger is not None:
            logger.info(
                "[OrtGradientForwardBackwardOptimizer.fit] "
                "end loss=%r", self.train_losses_[-1])
        return self

    def _iteration(self, data_loader, states, n_weights):
        actual_losses = []
        bs = data_loader.batch_size
        logger = self._logger
        if len(states) == 1:
            state = states[0]
            grad = None
        else:
            state, grad = states

        if logger is not None:
            logger.debug(
                "[OrtGradientForwardBackwardOptimizer._iteration] "
                "iteration begin learning_rate=%r",
                self.learning_rate)

        prediction_cache = None
        prediction_cache_shape = None
        backward_outputs_cache = None
        for ib, ito in enumerate(data_loader.iter_ortvalue()):
            if len(ito) == 2:
                (ortx, orty) = ito
                ortw = None
            else:
                (ortx, orty, ortw) = ito
            state[0] = ortx

            if logger is not None:
                logger.debug(
                    "[OrtGradientForwardBackwardOptimizer._iteration] "
                    "batch %d", ib)

            ortx_shape = tuple(ortx.shape())
            same_shape = (
                prediction_cache_shape is not None and
                ortx_shape == prediction_cache_shape)

            if logger is not None:
                logger.debug(
                    "[OrtGradientForwardBackwardOptimizer._iteration] forward")

            # forward
            if prediction_cache_shape is None or same_shape:
                prediction_cache = None
                prediction_cache_shape = None
            prediction = self.train_function_.forward(
                states[0], training=True,
                forward_outputs_cache=prediction_cache)
            prediction_cache = prediction
            prediction_cache_shape = ortx_shape

            if logger is not None:
                logger.debug(
                    "[OrtGradientForwardBackwardOptimizer._iteration] "
                    "loss types=%r,%r",
                    orty.data_type(), prediction[0].data_type())

            # loss
            loss, loss_gradient = self.learning_loss.loss_gradient(
                self.device, orty, prediction[0], weight=ortw)

            if logger is not None:
                logger.debug(
                    "[OrtGradientForwardBackwardOptimizer._iteration] "
                    "loss=%g has_weight=%r",
                    loss.numpy(), ortw is not None)

            n = len(state) - n_weights
            loss = self.learning_penalty.penalty_loss(
                self.device, loss, *state[n:])

            cpu_loss = loss.numpy()

            if logger is not None:
                logger.debug(
                    "[OrtGradientForwardBackwardOptimizer._iteration] "
                    "cpu_loss=%r", cpu_loss)

            if numpy.isinf(cpu_loss) or numpy.isnan(cpu_loss):
                raise ConvergenceError(
                    "Loss is nan, learning_rate=%r, "
                    "the gradient descent has failed "
                    "(past losses=%r)." % (
                        self.learning_rate,
                        [float(v) for v in (
                            actual_losses if len(actual_losses) < 5
                            else actual_losses[-5:])]))

            # backward
            if not same_shape:
                backward_outputs_cache = None
            gradient = self.train_function_.backward(
                [loss_gradient], backward_outputs_cache=backward_outputs_cache)
            backward_outputs_cache = gradient

            if len(gradient) != len(state):
                raise RuntimeError(  # pragma: no cover
                    "gradient and state should have the same length but "
                    "%r != %r." % (len(gradient), len(state)))

            n = len(state) - n_weights

            for i in range(n, len(state)):
                self.learning_penalty.update_weights(
                    i - n, self.device, state[i])
                self.learning_rate.update_weights(
                    i - n, self.device, state[i],
                    gradient[i], bs,
                    None if grad is None else grad[i])

            if logger is not None:
                logger.debug(
                    "[OrtGradientForwardBackwardOptimizer._iteration] "
                    "loss=%g n_weights=%d", cpu_loss, n)
                for i in range(n, len(state)):
                    logger.debug(
                        "[OrtGradientForwardBackwardOptimizer._iteration] "
                        "state[%i]=%s", i, str_ortvalue(state[i]))

            actual_losses.append(cpu_loss / bs)

        if logger is not None:
            logger.debug(
                "[OrtGradientForwardBackwardOptimizer._iteration] "
                "iteration end")

        return numpy.array(actual_losses).mean()

    def _evaluation(self, data_loader, state):
        bs = data_loader.batch_size
        logger = self._logger
        actual_losses = []
        for ib, (ortx, orty) in enumerate(data_loader.iter_ortvalue()):
            state[0] = ortx

            if logger is not None:
                logger.debug(  # pragma: no cover
                    "[OrtGradientForwardBackwardOptimizer._evaluation] "
                    "batch %d", ib)

            prediction = self.train_function_.forward(state, training=False)
            loss, _ = self.learning_loss.loss_gradient(
                self.device, orty, prediction[0])
            cpu_loss = loss.numpy()
            if numpy.isinf(cpu_loss) or numpy.isnan(cpu_loss):
                raise ConvergenceError(  # pragma: no cover
                    "Loss is nan, "
                    "the evaluation has failed "
                    "(past losses=%r)." %
                    [float(v) for v in (
                        actual_losses if len(actual_losses) < 5
                        else actual_losses[-5:])])
            actual_losses.append(cpu_loss / bs)

        return numpy.array(actual_losses).sum() / len(data_loader)

    def _create_training_session(
            self, model_onnx, weights_to_train, device):

        forback = OrtGradientForwardBackward(
            model_onnx, weights_to_train=weights_to_train,
            debug=False, enable_logging=False,
            providers=device_to_providers(device))
        inst = forback.new_instance()
        return (forback, inst)



"""
@file
@brief Rewrites operator in ONNX graph.
"""
from onnx.helper import make_graph
from onnx import NodeProto
from onnx.numpy_helper import to_array, from_array


def _unique_name(existing_names, name):
    """
    Returns a name different from any name in *existing_names*.

    :param existing_names: set of names
    :param name: current
    :return: unique name
    """
    if name not in existing_names:
        existing_names.add(name)
        return name
    name0 = name
    i = 2
    while name in existing_names:
        name = "%s_%d" % (name0, i)
        i += 1
    existing_names.add(name)
    return name


def _onnx_rewrite_operator_node(existing_names, node, sub_onx):
    """
    Replaces a node by a subgraph.

    :param existing_names: existing results names
    :param node: onnx node to replace
    :param sub_onx: onnx sub_graph to use as a replacement
    :return: new_initializer, new_nodes
    """
    if len(node.input) != len(sub_onx.graph.input):
        raise ValueError(  # pragma: no cover
            "Mismatch with the number of inputs for operator type %r. "
            "%d != %d." % (
                node.op_type, len(node.input), len(sub_onx.graph.nput)))
    if len(node.output) != len(sub_onx.graph.output):
        raise ValueError(  # pragma: no cover
            "Mismatch with the number of outputs for operator type %r. "
            "%d != %d." % (
                node.op_type, len(node.output), len(sub_onx.graph.output)))
    replaces = {}
    for inp, name in zip(sub_onx.graph.input, node.input):
        replaces[inp.name] = name
    for inp, name in zip(sub_onx.graph.output, node.output):
        replaces[inp.name] = name

    new_inits = []
    for init in sub_onx.graph.initializer:
        name = _unique_name(existing_names, init.name)
        replaces[init.name] = name
        tensor = from_array(to_array(init), name=name)
        new_inits.append(tensor)

    new_nodes = []
    for n in sub_onx.graph.node:
        new_node = NodeProto()
        new_node.op_type = n.op_type
        new_node.attribute.extend(n.attribute)  # pylint: disable=E1101
        new_node.input.extend(  # pylint: disable=E1101
            [replaces[i] for i in n.input])  # pylint: disable=E1101
        new_node.domain = n.domain
        new_out = []
        for o in n.output:
            if o in replaces:
                new_out.append(replaces[o])
            else:
                n = _unique_name(existing_names, o)
                new_out.append(n)
        new_node.output.extend(new_out)  # pylint: disable=E1101
        new_nodes.append(new_node)

    return new_inits, new_nodes


def onnx_rewrite_operator(onx, op_type, sub_onx, recursive=True, debug_info=None):
    """
    Replaces one operator by an onnx graph.

    :param onx: onnx graph
    :param op_type: operator type
    :param sub_onx: onnx graph
    :param recursive: looks into subgraphs
    :param debug_info: unused
    :return: modified onnx graph

    .. runpython::
        :showcode:

        import numpy
        from skl2onnx.common.data_types import FloatTensorType
        from skl2onnx.algebra.onnx_ops import (  # pylint: disable=E0611
            OnnxReciprocal, OnnxDiv)
        from mlprodict.plotting.text_plot import onnx_simple_text_plot
        from onnxcustom import get_max_opset
        from onnxcustom.utils.onnx_rewriter import onnx_rewrite_operator

        # first graph: it contains the node to replace
        opset = get_max_opset()
        node1 = OnnxReciprocal('X', output_names=['Y'],
                               op_version=opset)
        onx1 = node1.to_onnx(
            inputs={'X': FloatTensorType()},
            outputs={'Y': FloatTensorType()},
            target_opset=opset)

        # second graph: it contains the replacement graph
        node2 = OnnxDiv(numpy.array([1], dtype=numpy.float32),
                        'X', output_names=['Y'],
                        op_version=opset)
        onx2 = node2.to_onnx(
            inputs={'X': FloatTensorType()},
            outputs={'Y': FloatTensorType()},
            target_opset=opset)

        # third graph: the modified graph
        onx3 = onnx_rewrite_operator(onx1, 'Reciprocal', onx2)
        print(onnx_simple_text_plot(onx3))
    """
    from mlprodict.onnx_tools.optim._onnx_optimisation_common import (  # pylint: disable=C0415
        _apply_remove_node_fct_node, _apply_optimisation_on_graph)

    if hasattr(onx, 'graph'):
        fct = (lambda graph, recursive=False, debug_info=None:
               onnx_rewrite_operator(
                   graph, op_type, sub_onx, recursive=recursive,
                   debug_info=debug_info))
        return _apply_optimisation_on_graph(fct, onx, recursive=recursive)

    existing_names = set()
    for node in onx.node:
        existing_names.update(node.input)
        existing_names.update(node.output)

    nodes = list(onx.node)
    new_nodes = []
    new_inits = list(onx.initializer)
    for i, node in enumerate(nodes):
        if node.op_type != op_type:
            new_nodes.append(node)
            continue
        inits, newn = _onnx_rewrite_operator_node(
            existing_names, node, sub_onx)
        new_inits.extend(inits)
        new_nodes.extend(newn)

    if recursive:
        # Handles subgraphs.
        for i in range(len(new_nodes)):  # pylint: disable=C0200
            node = nodes[i]
            if node is None or not (node.attribute):  # pylint: disable=C0325
                continue
            nodes[i] = _apply_remove_node_fct_node(
                onnx_rewrite_operator, node, recursive=True,
                debug_info=None)

    graph = make_graph(
        new_nodes, onx.name, onx.input, onx.output, new_inits)
    return graph



"""
@file
@brief Command lines for profiling.
"""
from ..utils.nvprof2json import convert_trace_to_json


def nvprof2json(filename, output="", temporary_file="", verbose=1,
                fLOG=print):
    """
    Converts traces produced by :epkg:`nvprof` and saved with
    format :epkg:`sqlite3` (extension `.sql`).

    :param filename: filename
    :param output: output file, if left empty, the result is printed
        on the standard output
    :param temporary_file: if the file needs to be unzipped,
        this file will be created to be the unzipped file,
        it is not cleaned after the unzipping.
    :param verbose: verbosity
    :param fLOG: logging function
    :return: json (if output is None, the list of events otherwise)

    .. cmdref::
        :title: Converts a profile stored by nvprof into json
        :cmd: -m onnxcustom nvprof2json --help

        The sqlite dump is generated with a command line similar to:

        ::

            nvprof -o gpu_profile.sql python plot_gpu_training.py

        The command produces a json file following the *Trace Event Format*.
    """
    verbose = int(verbose)
    res = convert_trace_to_json(filename, output, verbose=verbose,
                                temporary_file=temporary_file, fLOG=fLOG)
    if output is None:  # pragma: no cover
        fLOG(res)



# coding: utf-8
# flake8: noqa: F401
# pylint: disable=W0611,C0415
"""
@file
@brief Experimentation with ONNX, examples.
"""

__version__ = "0.4.274"
__author__ = "Xavier Dupré, ..."
__max_supported_opset__ = 15  # Converters are tested up to this version.


def check(verbose=1):
    """
    Runs a couple of functions to check the module is working.

    :param verbose: 0 to hide the standout output
    :return: list of dictionaries, result of each test
    """
    tests = []
    try:
        import onnx
        import onnx.helper
    except ImportError as e:  # pragma: no cover
        tests.append(dict(test='onnx', exc=e))
    try:
        import onnxruntime
        from onnxruntime import InferenceSession
    except ImportError as e:  # pragma: no cover
        tests.append(dict(test='onnxruntime', exc=e))
    try:
        import onnxruntime.training
        from onnxruntime.training import TrainingSession
    except ImportError as e:  # pragma: no cover
        tests.append(dict(test='onnxruntime_training', exc=e))
    try:
        import skl2onnx
        from skl2onnx import to_onnx
    except ImportError as e:  # pragma: no cover
        tests.append(dict(test='skl2onnx', exc=e))
    return tests


def get_max_opset():
    """
    Returns the highest available onnx opset version.
    """
    from onnx.defs import onnx_opset_version
    return min(onnx_opset_version(), __max_supported_opset__)



# pylint: disable=W0105
"""
@file
@brief Helper for :epkg:`onnxruntime-training`.
"""
from onnxruntime import SessionOptions, InferenceSession, RunOptions
from ..utils.onnx_function import function_onnx_graph
from ..utils.onnxruntime_helper import device_to_providers
from ._base_onnx_function import BaseLearningOnnx


class BaseLearningPenalty(BaseLearningOnnx):
    """
    Class handling the penalty on the coefficients for class
    @see cl OrtGradientForwardBackwardOptimizer.
    """

    def __init__(self):
        BaseLearningOnnx.__init__(self)
        self.ro_ = RunOptions()

    def _call_iobinding(self, sess, bind):
        sess.run_with_iobinding(bind, self.ro_)

    @staticmethod
    def select(class_name, **kwargs):
        """
        Returns an instance of a given initialized with
        *kwargs*.
        :param class_name: an instance of @see cl BaseLearningPenalty
            or a string among the following class names (see below)
        :return: instance of @see cl BaseLearningPenalty

        Possible values for *class_name*:
        * None or `'penalty'`: see @see cl L1L2PenaltyLearning
        """
        if isinstance(class_name, BaseLearningPenalty):
            return class_name
        cls = {NoLearningPenalty: [None, ''],
               ElasticLearningPenalty: ['elastic', 'l1l2']}
        for cl, aliases in cls.items():
            if class_name == cl.__class__.__name__ or class_name in aliases:
                return cl(**kwargs)
        raise ValueError(  # pragma: no cover
            "Unexpected class name %r. It should be one of %r." % (
                class_name, list(map(lambda c: c.__name__, cls))))

    def penalty_loss(self, device, loss, *weights):
        """
        Returns the received loss. Updates the loss inplace.

        :param device: device where the training takes place
        :param loss: loss without penalty
        :param weights: any weights to be penalized
        :return: loss
        """
        raise NotImplementedError(
            "penalty_loss must be overwritten.")

    def update_weights(self, device, statei):
        """
        Returns the received loss. Updates the weight inplace.

        :param device: device where the training takes place
        :param statei: loss without penalty
        :return: weight
        """
        raise NotImplementedError(
            "update_weights must be overwritten.")


class NoLearningPenalty(BaseLearningPenalty):
    """
    No regularization.
    """

    def __init__(self):
        BaseLearningPenalty.__init__(self)

    def build_onnx_function(self, opset, device, n_tensors):
        # Nothing to do.
        pass

    def penalty_loss(self, device, loss, *weights):
        """
        Returns the received loss. Updates the loss inplace.

        :param device: device where the training takes place
        :param loss: loss without penalty
        :param weights: any weights to be penalized
        :return: loss
        """
        return loss

    def update_weights(self, n_bind, device, statei):
        """
        Returns the received loss. Updates the weight inplace.

        :param device: device where the training takes place
        :param statei: loss without penalty
        :return: weight
        """
        return statei


class ElasticLearningPenalty(BaseLearningPenalty):
    """
    Implements a L1 or L2 regularization on weights.
    """

    def __init__(self, l1=0.5, l2=0.5):
        BaseLearningPenalty.__init__(self)
        self.l1 = l1
        self.l2 = l2

    def build_onnx_function(self, opset, device, n_tensors):
        so = SessionOptions()
        so.log_severity_level = 4

        # loss_grad
        self.penalty_onnx_ = function_onnx_graph(
            "n_penalty_elastic_error", target_opset=opset, n_tensors=n_tensors,
            loss_shape=None, l1_weight=self.l1, l2_weight=self.l2)
        self.penalty_sess_ = InferenceSession(
            self.penalty_onnx_.SerializeToString(), so,
            providers=device_to_providers(device))
        self.penalty_sess_bind_ = (
            self.penalty_sess_.io_binding()._iobinding)
        self.names_ = [i.name for i in self.penalty_onnx_.graph.input]

        # weight updates
        self.penalty_grad_onnx_ = function_onnx_graph(
            "update_penalty_elastic_error", target_opset=opset,
            l1=self.l1, l2=self.l2)
        self.penalty_grad_sess_ = InferenceSession(
            self.penalty_grad_onnx_.SerializeToString(), so,
            providers=device_to_providers(device))
        self.penalty_grad_sess_binds_ = [
            self.penalty_grad_sess_.io_binding()._iobinding
            for n in range(n_tensors)]

    def penalty_loss(self, device, *inputs):
        """
        Computes the penalty associated to every
        weights and adds them up to the loss.

        :param device: device where the training takes place
        :param inputs: loss without penalty and weights
        :return: loss + penatlies
        """
        if (not hasattr(self, "penalty_onnx_") or
                not hasattr(self, "penalty_sess_bind_")):
            raise RuntimeError(  # pragma: no cover
                "Attributes 'penalty_sess_bind_' or 'penalty_onnx_' is "
                "missing. Method 'build_onnx_function' has not been called.")
        if len(self.names_) != len(inputs):
            raise RuntimeError(  # pragma: no cover
                "Mismatched number of inputs: %d != %d." % (
                    len(self.names_), len(inputs)))

        for name, inp in zip(self.names_, inputs):
            self._bind_input_ortvalue(
                name, self.penalty_sess_bind_, inp, device, cache=True)
        self._bind_output_ortvalue(
            'Y', self.penalty_sess_bind_, inputs[0], cache=True)
        self._call_iobinding(self.penalty_sess_._sess, self.penalty_sess_bind_)
        return self.penalty_sess_bind_.get_outputs()[0]

    def update_weights(self, n_bind, device, statei):
        if (not hasattr(self, "penalty_grad_onnx_") or
                not hasattr(self, "penalty_grad_sess_binds_")):
            raise RuntimeError(  # pragma: no cover
                "Attributes 'penalty_grad_sess_binds_' or "
                "'penalty_grad_onnx_' is missing. Method "
                "'build_onnx_function' has not been called.")
        bind = self.penalty_grad_sess_binds_[n_bind]
        self._bind_input_ortvalue("X", bind, statei, device, cache=True)
        self._bind_output_ortvalue('Y', bind, statei, cache=True)
        self._call_iobinding(self.penalty_grad_sess_._sess, bind)
        return bind.get_outputs()[0]  # X



"""
@file
@brief Helpers to improve documentation rendering.
"""
import re


def fix_link_operator_md(markdown):
    """
    The redering of file `Operator.md <https://github.com/onnx/onnx/
    blob/master/docs/Operators.md>`_ breaks links. This function
    restores some of them.

    :param markdown: a string or a filename
    :return: modified content
    """
    if len(markdown) < 5000 and markdown.endswith('.md'):
        with open(markdown, 'r', encoding='utf-8') as f:
            content = f.read()
    else:
        content = markdown  # pragma: no cover

    reg = re.compile(
        "([|]<a href=\\\"#(?P<name>[.A-Za-z]+)\\\">(?P=name)</a>[|])")
    pattern = "|[{1}](#a-name-{0}-a-a-name-{0}-{0}-a)|"

    lines = content.split('\n')
    new_lines = []
    for line in lines:
        find = reg.search(line)
        if find:
            gr = find.groups()
            exp = gr[0]
            op = gr[1]
            rep = pattern.format(op.lower(), op).replace(".", "-")
            line = line.replace(exp, rep)
        new_lines.append(line)
    return "\n".join(new_lines)



# flake8: noqa: F401
"""
@file
@brief Shortcuts to plotting.
"""
from mlprodict.plotting.plotting_onnx import plot_onnx


def plot_onnxs(*onx, ax=None, dpi=300, temp_dot=None, temp_img=None,
               show=False, title=None):
    """
    Plots one or several ONNX graph into a :epkg:`matplotlib` graph.

    :param onx: ONNX objects
    :param ax: existing axes
    :param dpi: resolution
    :param temp_dot: temporary file,
        if None, a file is created and removed
    :param temp_img: temporary image,
        if None, a file is created and removed
    :param show: calls `plt.show()`
    :param title: graph title
    :return: axes
    """
    if len(onx) == 1:
        if ax is None:
            import matplotlib.pyplot as plt  # pylint: disable=C0415
            ax = plt.gca()
        elif isinstance(ax, str) and ax == 'new':
            import matplotlib.pyplot as plt  # pylint: disable=C0415
            _, ax = plt.subplots(1, 1)
        ax = plot_onnx(onx[0], ax=ax, dpi=dpi, temp_dot=temp_dot,
                       temp_img=temp_img)
        if title is not None:
            ax.set_title(title)
        return ax
    elif len(onx) > 1 and isinstance(ax, str) and ax == 'new':
        ax = None

    if len(onx) == 0:
        raise ValueError(
            "Empty list of graph to plot.")

    if ax is None:
        import matplotlib.pyplot as plt  # pylint: disable=C0415
        fig, ax = plt.subplots(1, len(onx))
    else:
        fig = None
    if ax.shape[0] != len(onx):
        raise ValueError(
            "ax must be an array of shape (%d, )." % len(onx))
    for i, ox in enumerate(onx):
        plot_onnx(ox, ax=ax[i], dpi=dpi, temp_dot=temp_dot,
                  temp_img=temp_img)
        if title is None or isinstance(title, str):
            continue
        if i < len(title):
            ax[i].set_title(title[i])
    if len(onx) > 1 and isinstance(title, str):
        if fig is None:
            raise ValueError(  # pragma: no cover
                "Main title cannot be set if fig is undefined (title=%r, "
                "len(onx)=%d)" % (title, len(onx)))
        fig.suptitle(title)
    elif len(onx) == 1:
        if isinstance(title, list):
            title = title[0]
        ax.set_title(title)
    return ax
